from baconian.config.global_config import GlobalConfig
from baconian.test.tests.set_up.setup import TestWithLogSet
import numpy as np
from baconian.core.parameters import Parameters
from baconian.common.schedules import LinearScheduler, PiecewiseScheduler, PeriodicalEventSchedule

x = 0


class TestParam(TestWithLogSet):
    def test_basic(self):
        a, locals = self.create_parameters()
        a.save(save_path=GlobalConfig().DEFAULT_LOG_PATH + '/param_path',
               name=a.name,
               global_step=0)
        or_val = a._source_config.config_dict['var1']
        or_param = a('param3').copy()
        a._source_config.config_dict['var1'] = 100
        a._parameters['param3'] = 1000
        self.assertNotEqual(a._source_config.config_dict['var1'], or_val)
        self.assertFalse(np.equal(a._parameters['param3'], or_param).all())
        a.load(load_path=GlobalConfig().DEFAULT_LOG_PATH + '/param_path',
               name=a.name,
               global_step=0
               )
        self.assertEqual(a._source_config.config_dict['var1'], or_val)
        self.assertTrue(np.equal(a._parameters['param3'], or_param).all())
        b, _ = self.create_parameters()
        b.copy_from(a)

        for key in a._source_config.required_key_dict.keys():
            if isinstance(a[key], np.ndarray):
                self.assertTrue(np.equal(a[key], b[key]).all())
            else:
                self.assertEqual(id(a[key]), id(b[key]))
                self.assertEqual(id(a(key)), id(b(key)))
        for key in a._parameters.keys():
            if isinstance(a[key], np.ndarray):
                self.assertTrue(np.equal(a[key], b[key]).all())
            else:
                self.assertEqual(a[key], b[key])
                self.assertEqual(a(key), b(key))

    def test_scheduler_param(self):
        def func():
            global x
            return x

        parameters = dict(param1='aaaa',
                          param2=1.0,
                          param4=1.0,
                          param3=np.random.random([4, 2]))
        source_config, _ = self.create_dict_config()
        a = Parameters(parameters=parameters,
                       source_config=source_config,
                       name='test_params',
                       to_scheduler_param_tuple=(dict(param_key='param2',
                                                      scheduler=LinearScheduler(t_fn=func,
                                                                                schedule_timesteps=10,
                                                                                final_p=0.0)),
                                                 dict(param_key='param4',
                                                      scheduler=PiecewiseScheduler(t_fn=func,
                                                                                   endpoints=(
                                                                                      (2, 0.5), (8, 0.2), (10, 0.0)),
                                                                                   outside_value=0.0,
                                                                                   ))))
        a.init()
        for i in range(20):
            global x
            if x < 10:
                self.assertEqual(a('param2'), 1.0 - x * (1.0 - 0.0) / 10)
            else:
                self.assertEqual(a('param2'), 0.0)
            if x == 2:
                self.assertEqual(a('param4'), 0.5)
            if x == 8:
                self.assertEqual(a('param4'), 0.2)
            if x >= 10:
                self.assertEqual(a('param4'), 0.0)
            x += 1
        b, _ = self.create_parameters()
        b.copy_from(a)
        for key in a._source_config.required_key_dict.keys():
            if isinstance(a[key], np.ndarray):
                self.assertTrue(np.equal(a[key], b[key]).all())
            else:
                self.assertEqual(id(a[key]), id(b[key]))
                self.assertEqual(id(a(key)), id(b(key)))
        for key in a._parameters.keys():
            if isinstance(a[key], np.ndarray):
                self.assertTrue(np.equal(a[key], b[key]).all())
            else:
                self.assertEqual(a[key], b[key])
                self.assertEqual(a(key), b(key))
        self.assertEqual(a.to_scheduler_param_list.__len__(), b.to_scheduler_param_list.__len__())
        for a_val, b_val in zip(a.to_scheduler_param_list, b.to_scheduler_param_list):
            self.assertEqual(a_val['param_key'], b_val['param_key'])
            self.assertEqual(a_val['scheduler'].value(), b_val['scheduler'].value())

    def test_event_schedule(self):
        def func():
            global x
            return x

        a = PeriodicalEventSchedule(t_fn=func,
                                    trigger_every_step=5,
                                    after_t=10)
        for i in range(100):
            if i % 5 != 0 or i < 10:
                self.assertFalse(a.value())
            else:
                self.assertTrue(a.value())
            global x
            x += 1



import os
import random
import time

import numpy as np
import tensorflow as tf
from GPUtil import GPUtil as Gpu
from typeguard import typechecked

from baconian.common import files as file
from baconian.common.logging import Logger, ConsoleLogger
from baconian.config.global_config import GlobalConfig
from copy import deepcopy
import tracemalloc


def _reset_global_seed(seed):
    """

    :param seed: seed to reset global variables
    :type seed: int
    """
    sess = tf.get_default_session()
    if sess:

        sess.__exit__(None, None, None)
    tf.reset_default_graph()
    tf.set_random_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


@typechecked
def single_exp_runner(task_fn, auto_choose_gpu_flag=False, gpu_id: int = 0, seed=None, del_if_log_path_existed=False,
                      **task_fn_kwargs):

    """

    :param task_fn: task function defined bu users
    :type task_fn: method
    :param auto_choose_gpu_flag: auto choose gpu, default False
    :type auto_choose_gpu_flag: bool
    :param gpu_id: gpu id, default 0
    :type gpu_id: int
    :param seed: seed generated by system time
    :type seed: int
    :param del_if_log_path_existed:delete obsolete log file path if existed, by default False
    :type del_if_log_path_existed: bool
    :param task_fn_kwargs:
    :type task_fn_kwargs:
    :return:
    :rtype:
    """
    os.environ['CUDA_DEVICE_ORDER'] = "PCI_BUS_ID"
    if auto_choose_gpu_flag is True:
        DEVICE_ID_LIST = Gpu.getFirstAvailable()
        DEVICE_ID = DEVICE_ID_LIST[0]
        os.environ["CUDA_VISIBLE_DEVICES"] = str(DEVICE_ID)
    else:
        os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_id)
    if not seed:
        seed = int(round(time.time() * 1000)) % (2 ** 32 - 1)
    _reset_global_seed(seed)
    print("create log path at {}".format(GlobalConfig().DEFAULT_LOG_PATH), flush=True)

    file.create_path(path=GlobalConfig().DEFAULT_LOG_PATH, del_if_existed=del_if_log_path_existed)
    Logger().init(config_or_config_dict=dict(),
                  log_path=GlobalConfig().DEFAULT_LOG_PATH,
                  log_level=GlobalConfig().DEFAULT_LOG_LEVEL)
    ConsoleLogger().init(to_file_flag=GlobalConfig().DEFAULT_WRITE_CONSOLE_LOG_TO_FILE_FLAG,
                         to_file_name=os.path.join(GlobalConfig().DEFAULT_LOG_PATH,
                                                   GlobalConfig().DEFAULT_CONSOLE_LOG_FILE_NAME),
                         level=GlobalConfig().DEFAULT_LOG_LEVEL,
                         logger_name=GlobalConfig().DEFAULT_CONSOLE_LOGGER_NAME)

    task_fn(**task_fn_kwargs)


@typechecked
def duplicate_exp_runner(num, task_fn, auto_choose_gpu_flag=False, gpu_id: int = 0, seeds: list = None,
                         del_if_log_path_existed=False,
                         **task_fn_kwargs):

    """

    :param num: the number of multiple experiments
    :type num: int
    :param task_fn: task function, defined by users
    :type task_fn: method
    :param auto_choose_gpu_flag: auto choose gpu, default False
    :type auto_choose_gpu_flag: bool
    :param gpu_id: gpu id, default 0
    :type gpu_id: int
    :param seeds: seeds generated by system time
    :type seeds: list
    :param del_if_log_path_existed: delete the existing log path, default False
    :type del_if_log_path_existed: bool
    :param task_fn_kwargs:
    :type task_fn_kwargs:
    :return:
    :rtype:
    """
    # tracemalloc.start(100)
    if seeds:
        assert len(seeds) == num
    base_log_path = deepcopy(GlobalConfig().DEFAULT_LOG_PATH)

    for i in range(num):
        GlobalConfig().set('DEFAULT_LOG_PATH', os.path.join(base_log_path, 'exp_{}'.format(i)))
        single_exp_runner(task_fn=task_fn, auto_choose_gpu_flag=auto_choose_gpu_flag,
                          del_if_log_path_existed=del_if_log_path_existed,
                          gpu_id=gpu_id, seed=seeds[i] if seeds else None, **task_fn_kwargs)



from baconian.common.sampler.sample_data import TransitionData
from baconian.test.tests.set_up.setup import TestWithLogSet
import numpy as np
from baconian.algo.dynamics.terminal_func.terminal_func import *

x = 0


def func():
    return x


class TestRewardTerminalFunc(TestWithLogSet):

    def test_all_reward_func(self):
        pass

    def test_all_terminal_func(self):
        a = FixedEpisodeLengthTerminalFunc(max_step_length=10,
                                           step_count_fn=func)
        global x
        for i in range(11):
            if x == 10:
                self.assertTrue(a(state=None, action=None, new_state=None))
            else:
                self.assertFalse(a(state=None, action=None, new_state=None))
            x += 1



"""
A simple example to show how to wrap original environment's observation space, action space and reward function for
reshaping to boost the agent's training.
Actually, this is the feature supported by gym, but since we develop a new environment class based on gym's env, so
a tutorial is given th better introduce the usage of it.
"""

from baconian.envs.env_wrapper import ObservationWrapper, ActionWrapper, RewardWrapper
from baconian.envs.gym_env import make


class SmoothMountainCarReward(RewardWrapper):

    def _reward(self, observation, action, reward, done, info):
        return 10.0


car_env = make('MountainCarContinuous-v0')
car_env = SmoothMountainCarReward(env=car_env)
car_env.reset()
new_st, reward, terminal, info = car_env.step(action=car_env.action_space.sample())
print(reward)



from baconian.test.tests.set_up.setup import TestTensorflowSetup
import tensorflow as tf
from baconian.algo.dynamics.dynamics_model import DifferentiableDynamics
import numpy as np


class TestBasicClassInAlgo(TestTensorflowSetup):
    def test_derivable(self):
        val = 10.0
        val2 = 2.0
        bs = 1
        in_node = tf.placeholder(shape=[None, 10], dtype=tf.float32)
        in_node2 = tf.placeholder(shape=[None, 10], dtype=tf.float32)

        out_node = in_node * in_node * val + in_node2 * in_node2 * val2
        a = DifferentiableDynamics(input_node_dict=dict(in_node=in_node, in_node2=in_node2),
                                   output_node_dict=dict(out_node=out_node))
        self.sess.run(tf.global_variables_initializer())
        res = self.sess.run(a.grad_on_input(key_or_node='in_node'), feed_dict={
            in_node: np.random.random([bs, 10]),
            in_node2: np.random.random([bs, 10])
        })
        print('jacobian {}'.format(np.array(res).shape))

        # self.assertTrue(np.equal(res, val).all())

        res = self.sess.run(a.grad_on_input(key_or_node='in_node2'), feed_dict={
            in_node: np.random.random([bs, 10]),
            in_node2: np.random.random([bs, 10])
        })
        print('jacobian {}'.format(np.array(res).shape))

        # self.assertTrue(np.equal(res, val2).all())

        res = self.sess.run(a.grad_on_input(key_or_node=in_node), feed_dict={
            in_node: np.random.random([bs, 10]),
            in_node2: np.random.random([bs, 10])
        })
        # self.assertTrue(np.equal(res, val).all())

        res = self.sess.run(a.grad_on_input(key_or_node=in_node2), feed_dict={
            in_node: np.random.random([bs, 10]),
            in_node2: np.random.random([bs, 10])
        })
        # self.assertTrue(np.equal(res, val2).all())

        res = self.sess.run(a.grad_on_input(key_or_node=in_node, order=2), feed_dict={
            in_node: np.random.random([bs, 10]),
            in_node2: np.random.random([bs, 10])
        })
        print('hessian {}'.format(np.array(res).shape))

        # self.assertTrue(np.equal(res, 0).all())

        res = self.sess.run(a.grad_on_input(key_or_node=in_node2, order=2), feed_dict={
            in_node: np.random.random([bs, 10]),
            in_node2: np.random.random([bs, 10])
        })
        print('hessian {}'.format(np.array(res).shape))

        # self.assertTrue(np.equal(res, 0).all())



import numpy as np
from matplotlib import pyplot as plt
import json
import seaborn as sns

sns.set_style('whitegrid')


class Plotter(object):
    markers = ('+', 'x', 'v', 'o', '^', '<', '>', 's', 'p', '*', 'h', 'H', 'D', 'd', 'P', 'X')
    color_list = ['b', 'r', 'g', 'm', 'y', 'k', 'cyan', 'plum', 'darkgreen', 'darkorange', 'oldlace', 'chocolate',
                  'purple', 'lightskyblue', 'gray', 'seagreen', 'antiquewhite',
                  'snow', 'darkviolet', 'brown', 'skyblue', 'mediumaquamarine', 'midnightblue', 'darkturquoise',
                  'sienna', 'lightsteelblue', 'gold', 'teal', 'blueviolet', 'mistyrose', 'seashell', 'goldenrod',
                  'forestgreen', 'aquamarine', 'linen', 'deeppink', 'darkslategray', 'mediumseagreen', 'dimgray',
                  'mediumpurple', 'lightgray', 'khaki', 'dodgerblue', 'papayawhip', 'salmon', 'floralwhite',
                  'lightpink', 'gainsboro', 'coral', 'indigo', 'darksalmon', 'royalblue', 'navy', 'orangered',
                  'cadetblue', 'orchid', 'palegreen', 'magenta', 'honeydew', 'darkgray', 'palegoldenrod', 'springgreen',
                  'lawngreen', 'palevioletred', 'olive', 'red', 'lime', 'yellowgreen', 'aliceblue', 'orange',
                  'chartreuse', 'lavender', 'paleturquoise', 'blue', 'azure', 'yellow', 'aqua', 'mediumspringgreen',
                  'cornsilk', 'lightblue', 'steelblue', 'violet', 'sandybrown', 'wheat', 'greenyellow', 'darkred',
                  'mediumslateblue', 'lightseagreen', 'darkblue', 'moccasin', 'lightyellow', 'turquoise', 'tan',
                  'mediumvioletred', 'mediumturquoise', 'limegreen', 'slategray', 'lightslategray', 'mintcream',
                  'darkgreen', 'white', 'mediumorchid', 'firebrick', 'bisque', 'darkcyan', 'ghostwhite', 'powderblue',
                  'tomato', 'lavenderblush', 'darkorchid', 'cornflowerblue', 'plum', 'ivory', 'darkgoldenrod', 'green',
                  'burlywood', 'hotpink', 'cyan', 'silver', 'peru', 'thistle', 'indianred', 'olivedrab',
                  'lightgoldenrodyellow', 'maroon', 'black', 'crimson', 'darkolivegreen', 'lightgreen', 'darkseagreen',
                  'lightcyan', 'saddlebrown', 'deepskyblue', 'slateblue', 'whitesmoke', 'pink', 'darkmagenta',
                  'darkkhaki', 'mediumblue', 'beige', 'blanchedalmond', 'lightsalmon', 'lemonchiffon', 'navajowhite',
                  'darkslateblue', 'lightcoral', 'rosybrown', 'fuchsia', 'peachpuff']

    def plot_fig(self, fig_num, col_id, x, y, title, x_label, y_label, label=' ', marker='*'):
        plt.figure(fig_num, figsize=(6, 5))
        plt.title(title)
        plt.xlabel(x_label)
        plt.ylabel(y_label)
        plt.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
        plt.tight_layout()

        marker_every = max(int(len(x) / 10), 1)
        if len(np.array(y).shape) > 1:
            new_shape = np.array(y).shape

            res = np.reshape(np.reshape(np.array([y]), newshape=[-1]), newshape=[new_shape[1], new_shape[0]],
                             order='F').tolist()
            res = list(res)
            for i in range(len(res)):
                res_i = res[i]
                plt.subplot(len(res), 1, i + 1)
                plt.title(title + '_' + str(i))
                plt.plot(x, res_i, self.color_list[col_id], label=label + '_' + str(i), marker=marker,
                         markevery=marker_every, markersize=6, linewidth=1)
                col_id += 1
        else:
            plt.plot(x, y, self.color_list[col_id], label=label, marker=marker, markevery=marker_every, markersize=6,
                     linewidth=1)
        plt.legend()

    @staticmethod
    def plot_any_key_in_log(data, key, index, exp_num=1,
                            sub_log_dir_name=None,
                            scatter_flag=False,
                            histogram_flag=False,
                            save_flag=False,
                            save_path=None,
                            save_format='png',
                            file_name=None,
                            separate_exp_flag=True,
                            mean_stddev_flag=False,
                            path_list=None,
                            ):
        """
        :param data: a pandas DataFrame containing (index and) key columns
        :param key: in y-axis, the variable to plot, assigned by user
        :param index: in x-axis, the argument assigned by user
        :param sub_log_dir_name: the sub-directory which the log file to plot is saved in
        :param exp_num: [optional] the number of experiments to visualize
        :param scatter_flag: [optional] draw scatter plot if true
        :param histogram_flag: [optional] draw histogram if true
        :param save_flag: [optional] save the figure to a file if true
        :param save_path: [optional] save path of figure, assigned by user, the directory of log_path by default
        :param save_format: [optional] format of figure to save, png by default
        :param file_name: [optional] the file name of the file to save, key_VERUS_index by default
        :param separate_exp_flag: [optional] plot the results of each experiment separately if true
        :param mean_stddev_flag: [optional] plot the mean value of multiple experiment results and standard deviation
        :param path_list: [optional] the list of save paths assigned by users, figure file will be saved to each path
        :return:
        """

        marker_every = max(int(data.shape[0] / 10), 1)

        # plt.figure()
        fig, ax = plt.subplots(1)

        if separate_exp_flag is True:
            for i in range(exp_num):
                if scatter_flag is True:
                    ax.scatter(data[index], data.iloc[:, i + 1], lw=1, label=key + '_' + str(i),
                               c=Plotter.color_list[i], alpha=0.8, )
                elif histogram_flag is True:
                    num_bins = 20
                    n, bins, patches = ax.hist(x=data[0], bins=num_bins)
                else:
                    ax.plot(data[index], data.iloc[:, i + 1], lw=1, label=key + '_' + str(i),
                            color=Plotter.color_list[i],
                            marker=Plotter.markers[i], markevery=marker_every, markersize=6, )
        if mean_stddev_flag is True:
            if histogram_flag is not True:
                ax.plot(data[index], data['MEAN'], lw=3, label='MEAN', color='silver')
                ax.fill_between(data[index], data['MEAN'] + data['STD_DEV'], data['MEAN'] - data['STD_DEV'],
                                facecolor='silver', alpha=0.5)

        plt.title(sub_log_dir_name)
        lgd = ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2), shadow=True, ncol=3)
        if histogram_flag is not True:
            plt.xlabel(index)
            plt.ylabel(key)
        else:
            plt.xlabel(key)
            plt.ylabel('count')
        plt.legend()
        # Save the figure to a file to a path or paths in a list
        if save_flag is True:
            if file_name is None:
                file_name = '/%s_VERSUS_%s' % (key, index)

            if path_list is not None:
                for path in path_list:
                    plt.savefig(path + '/%s.%s' % (file_name, save_format), bbox_extra_artists=(lgd,),
                                bbox_inches='tight', format=save_format)
                    print("Save plot figure to {path} as {file_name}".format(path=path,
                                                                             file_name='%s.%s' % (
                                                                                 file_name, save_format)))
            if save_path is not None:
                plt.savefig(save_path + '/%s.%s' % (file_name, save_format), bbox_extra_artists=(lgd,),
                            bbox_inches='tight', format=save_format)
                print("Save plot figure to {path} as {file_name}".format(path=save_path,
                                                                         file_name='%s.%s' % (
                                                                             file_name, save_format)))
        plt.show()

    @staticmethod
    def plot_any_scatter_in_log(res_dict, res_name, file_name, key, index, op, scatter_flag=False, save_flag=False,
                                save_path=None,
                                fig_id=4, label='', restrict_dict=None):
        with open(res_dict[res_name], 'r') as f:
            path_list = json.load(f)
        plt.figure(fig_id)
        plt.title("%s_%s_%s" % (res_name, file_name, key))
        plt.xlabel('index')
        plt.ylabel(key)
        for i in range(len(path_list)):
            test_reward = []
            real_env_sample_count_index = []
            with open(file=path_list[i] + '/loss/' + file_name, mode='r') as f:
                test_data = json.load(fp=f)
                for sample in test_data:
                    if restrict_dict is not None:
                        flag = True
                        for re_key, re_value in restrict_dict.items():
                            if sample[re_key] != re_value:
                                flag = False
                        if flag is True:
                            test_reward.append(sample[key])
                            real_env_sample_count_index.append(sample[index])
                    else:
                        test_reward.append(sample[key])
                        real_env_sample_count_index.append(sample[index])
            test_reward, real_env_sample_count_index = op(test_reward, real_env_sample_count_index)
            x_keys = []
            y_values = []
            last_key = real_env_sample_count_index[0]
            last_set = []

            for j in range(len(real_env_sample_count_index)):
                if real_env_sample_count_index[j] == last_key:
                    last_set.append(test_reward[j])
                else:
                    x_keys.append(last_key)
                    y_values.append(last_set)
                    last_key = real_env_sample_count_index[j]
                    last_set = [test_reward[j]]
            x_keys.append(last_key)
            y_values.append(last_set)
            y_values_mean = [np.mean(y_values[j]) for j in range(len(y_values))]
            if scatter_flag is True:
                plt.scatter(x_keys, y_values_mean, c=Plotter.color_list[i], label=key + label + str(i),
                            marker=Plotter.markers[i])
            else:
                plt.plot(x_keys, y_values_mean, c=Plotter.color_list[i], label=key + label + str(i),
                         marker=Plotter.markers[i])

        plt.legend()
        if save_flag is True:
            for path in path_list:
                plt.savefig(path + '/%s_%s.png' % (file_name, key))
        if save_path is not None:
            plt.savefig(save_path)
            print()
        plt.show()



from baconian.test.tests.set_up.setup import TestWithAll
from baconian.core.status import StatusCollector


class TestStatus(TestWithAll):

    def test_status_collector(self):
        a = StatusCollector()

        algo, local = self.create_dqn()
        env = local['env']
        env_spec = local['env_spec']
        agent, _ = self.create_agent(algo=algo, env=env,
                                     env_spec=env_spec,
                                     eps=self.create_eps(env_spec=env_spec)[0])
        self.register_global_status_when_test(env=env, agent=agent)

        agent.init()

        a.register_info_key_status(obj=agent, info_key='predict_counter', under_status='TRAIN',
                                   return_name='train_counter')
        a.register_info_key_status(obj=agent, info_key='predict_counter', under_status='TEST',
                                   return_name='test_counter')
        env.reset()
        agent.sample(env=env, sample_count=10, store_flag=True, in_which_status='TRAIN')
        agent.sample(env=env, sample_count=10, store_flag=True, in_which_status='TEST')
        agent.sample(env=env, sample_count=10, store_flag=True, in_which_status='TRAIN')

        res = a()
        self.assertTrue(len(res) == 2)

        self.assertTrue('train_counter' in res)
        self.assertTrue('test_counter' in res)

        self.assertTrue(res['test_counter'] == 10)
        self.assertTrue(res['train_counter'] == 20)


class TestStatusWithDQN(TestWithAll):
    def test_with_dqn(self):
        dqn, local = self.create_dqn()
        env = local['env']
        env_spec = local['env_spec']
        dqn.init()
        st = env.reset()
        from baconian.common.sampler.sample_data import TransitionData
        a = TransitionData(env_spec)
        res = []
        for i in range(100):
            ac = dqn.predict(obs=st, sess=self.sess, batch_flag=False)
            st_new, re, done, _ = env.step(action=ac)
            a.append(state=st, new_state=st_new, action=ac, done=done, reward=re)
            dqn.append_to_memory(a)
        res.append(dqn.train(batch_data=a, train_iter=10, sess=None, update_target=True)['average_loss'])
        res.append(dqn.train(batch_data=None, train_iter=10, sess=None, update_target=True)['average_loss'])
        print(dqn._status())
        print(dqn._status._info_dict_with_sub_info)



"""
The script to store some global configuration
"""

from typeguard import typechecked
import json_tricks as json
import tensorflow as tf
import os
from baconian.config.required_keys import SRC_UTIL_REQUIRED_KEYS
from baconian.common.error import *


class _SingletonDefaultGlobalConfig(object):
    DEFAULT_MAX_TF_SAVER_KEEP = 5
    DEFAULT_ALLOWED_EXCEPTION_OR_ERROR_LIST = (tf.errors.ResourceExhaustedError,)
    DEFAULT_BASIC_STATUS_LIST = ('TRAIN', 'TEST')
    DEFAULT_BASIC_INIT_STATUS = None

    # config required key list
    DEFAULT_DQN_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS, 'dqn.json')
    DEFAULT_CONSTANT_ACTION_POLICY_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS,
                                                                    'constant_action_policy.json')

    DEFAULT_ALGO_DYNA_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS,
                                                       'dyna.json')

    DEFAULT_ALGO_ME_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS,
                                                     'me.json')
    DEFAULT_MODEL_FREE_PIPELINE_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS,
                                                                 'model_free_pipeline.json')

    DEFAULT_MODEL_BASED_PIPELINE_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS,
                                                                  'model_based_pipeline.json')

    DEFAULT_MPC_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS,
                                                 'mpc.json')

    DEFAULT_DDPG_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS,
                                                  'ddpg.json')
    DEFAULT_MADDPG_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS,
                                                    'maddpg.json')

    DEFAULT_PPO_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS, 'ppo.json')
    DEFAULT_AGENT_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS, 'agent.json')

    DEFAULT_EXPERIMENT_REQUIRED_KEY_LIST = os.path.join(SRC_UTIL_REQUIRED_KEYS, 'experiment.json')

    # LOGGING CONFIG

    DEFAULT_ALLOWED_LOG_FILE_TYPES = ('json', 'csv', 'h5py')
    DEFAULT_LOG_LEVEL = 'DEBUG'
    from baconian import ROOT_PATH
    DEFAULT_LOG_PATH = os.path.join(ROOT_PATH, 'test/tests/tmp_path')
    DEFAULT_MODEL_CHECKPOINT_PATH = os.path.join(DEFAULT_LOG_PATH, 'model_checkpoints')
    DEFAULT_LOG_CONFIG_DICT = dict()
    DEFAULT_LOG_USE_GLOBAL_MEMO_FLAG = True

    DEFAULT_LOGGING_FORMAT = '%(levelname)s:%(asctime)-15s: %(message)s'
    DEFAULT_WRITE_CONSOLE_LOG_TO_FILE_FLAG = True
    DEFAULT_CONSOLE_LOG_FILE_NAME = 'console.log'
    DEFAULT_CONSOLE_LOGGER_NAME = 'console_logger'
    DEFAULT_EXPERIMENT_END_POINT = dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=500,
                                        TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                        TOTAL_AGENT_UPDATE_COUNT=None)

    DEFAULT_TURN_OFF_GLOBAL_NAME_FLAG = False

    # For internal use
    SAMPLE_TYPE_SAMPLE_TRANSITION_DATA = 'transition_data'
    SAMPLE_TYPE_SAMPLE_TRAJECTORY_DATA = 'trajectory_data'

    def __init__(self):
        super().__setattr__('freeze_flag', False)

    def freeze(self):
        super().__setattr__('freeze_flag', True)

    def unfreeze(self):
        super().__setattr__('freeze_flag', False)

    @typechecked
    def set_new_config(self, config_dict: dict):
        for key, val in config_dict.items():
            if hasattr(self, key):
                attr = getattr(self, key)
                if attr is not None and not isinstance(val, type(attr)):
                    raise TypeError('Set the GlobalConfig.{} with type{}, instead of type {}'.format(key,
                                                                                                     type(
                                                                                                         attr).__name__,
                                                                                                     type(
                                                                                                         val).__name__))
                setattr(self, key, val)
            else:
                setattr(self, key, val)

    @typechecked
    def set_new_config_by_file(self, path_to_file: str):
        with open(path_to_file, 'r') as f:
            new_dict = json.load(f)
            self.set_new_config(new_dict)

    @typechecked
    def set(self, key: str, val):
        if self.freeze_flag is True:
            raise AttemptToChangeFreezeGlobalConfigError()
        if hasattr(self, key):
            attr = getattr(self, key)
            if attr is not None and not isinstance(val, type(attr)):
                raise TypeError('Set the GlobalConfig.{} with type{}, instead of type {}'.format(key,
                                                                                                 type(
                                                                                                     attr).__name__,
                                                                                                 type(
                                                                                                     val).__name__))
            setattr(self, key, val)
            if key == 'DEFAULT_LOG_PATH':
                self.set('DEFAULT_MODEL_CHECKPOINT_PATH', os.path.join(val, 'model_checkpoints'))
        else:
            setattr(self, key, val)

    def return_all_as_dict(self):
        return_dict = {}
        for key in dir(self):
            if key.isupper() is True or 'DEFAULT' in key:
                attr = getattr(self, key)
                try:
                    json.dumps(dict(key=attr))
                except TypeError as e:
                    attr = 'cannot be json dumped'
                return_dict[key] = attr
        return return_dict

    def __setattr__(self, key, value):
        if self.freeze_flag is True:
            raise AttemptToChangeFreezeGlobalConfigError("{} {}".format(key, value))
        else:
            super().__setattr__(key, value)


class GlobalConfig(object):
    only_instance = None

    def __new__(cls, *args, **kwargs):
        if GlobalConfig.only_instance is None:
            GlobalConfig.only_instance = _SingletonDefaultGlobalConfig()
        return GlobalConfig.only_instance



from baconian.common.spaces.base import Space
import numpy as np
from typeguard import typechecked
from baconian.core.parameters import Parameters
from baconian.common.schedules import Scheduler


class ExplorationStrategy(object):
    def __init__(self):
        self.parameters = None

    def predict(self, **kwargs):
        raise NotImplementedError


class EpsilonGreedy(ExplorationStrategy):
    @typechecked
    def __init__(self, action_space: Space, init_random_prob: float, prob_scheduler: Scheduler = None):
        super(ExplorationStrategy, self).__init__()

        self.action_space = action_space
        self.random_prob_func = lambda: init_random_prob
        if prob_scheduler:
            self.random_prob_func = prob_scheduler.value

        self.parameters = Parameters(parameters=dict(random_prob_func=self.random_prob_func),
                                     name='eps_greedy_params')

    def predict(self, **kwargs):
        if np.random.random() < self.parameters('random_prob_func')():
            return self.action_space.sample()
        else:
            algo = kwargs.pop('algo')
            return algo.predict(**kwargs)



from baconian.test.tests.set_up.setup import TestWithAll
from baconian.common.sampler.sample_data import SampleData


class TestAgent(TestWithAll):
    def test_agent(self):
        algo, local = self.create_dqn()
        env = local['env']
        env_spec = local['env_spec']
        agent, _ = self.create_agent(algo=algo, env=env,
                                     env_spec=env_spec,
                                     eps=self.create_eps(env_spec=env_spec)[0])
        self.register_global_status_when_test(agent, env)
        agent.init()
        env.reset()
        data = agent.sample(env=env, sample_count=10, store_flag=False, in_which_status='TEST')
        self.assertTrue(isinstance(data, SampleData))
        self.assertEqual(agent.algo.replay_buffer.nb_entries, 0)
        data = agent.sample(env=env, sample_count=10, store_flag=True, in_which_status='TRAIN')
        self.assertTrue(isinstance(data, SampleData))
        self.assertEqual(agent.algo.replay_buffer.nb_entries, 10)

    def test_test(self):
        algo, local = self.create_dqn()
        env = local['env']
        env_spec = local['env_spec']
        agent, _ = self.create_agent(algo=algo, env=env,
                                     env_spec=env_spec,
                                     eps=self.create_eps(env_spec=env_spec)[0])
        self.register_global_status_when_test(agent, env)
        agent.init()
        env.reset()
        agent.test(sample_count=2)



from .replay_buffer import BaseReplayBuffer, UniformRandomReplayBuffer
from .epsilon_greedy import ExplorationStrategy, EpsilonGreedy
from .placeholder_input import PlaceholderInput, MultiPlaceholderInput
from .sample_processor import SampleProcessor



from baconian.test.tests.set_up.setup import BaseTestCase
from baconian.core.experiment_runner import single_exp_runner, duplicate_exp_runner
from baconian.common.schedules import LinearScheduler
from baconian.config.global_config import GlobalConfig
import os


class TestExperiment(BaseTestCase):
    def test_experiment(self):
        def func():
            GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT', dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=200,
                                                                    TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                                                    TOTAL_AGENT_UPDATE_COUNT=None))
            dqn, locals = self.create_dqn()
            env_spec = locals['env_spec']
            env = locals['env']
            agent = self.create_agent(env=locals['env'],
                                      algo=dqn,
                                      name='agent',
                                      eps=self.create_eps(env_spec)[0],
                                      env_spec=env_spec)[0]
            exp = self.create_exp(name='model_free', env=env, agent=agent)
            exp.run()

        single_exp_runner(func, auto_choose_gpu_flag=False, gpu_id=0, del_if_log_path_existed=True)

    def test_exp_with_scheduler(self, algo=None, locals=None):
        def wrap_algo(algo=None, locals=None):
            def func(algo=algo, locals=locals):
                GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT', dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=500,
                                                                        TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                                                        TOTAL_AGENT_UPDATE_COUNT=None))
                if not algo:
                    algo, locals = self.create_dqn()
                env_spec = locals['env_spec']
                env = locals['env']
                agent = self.create_agent(env=locals['env'],
                                          algo=algo,
                                          name='agent',
                                          eps=self.create_eps(env_spec)[0],
                                          env_spec=env_spec)[0]

                exp = self.create_exp(name='model_free', env=env, agent=agent)
                algo.parameters.set_scheduler(param_key='LEARNING_RATE',
                                              to_tf_ph_flag=True,
                                              scheduler=LinearScheduler(
                                                  t_fn=exp.TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT,
                                                  schedule_timesteps=GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[
                                                      'TOTAL_AGENT_TRAIN_SAMPLE_COUNT'],
                                                  final_p=0.0001,
                                                  initial_p=0.01))
                exp.run()
                self.assertEqual(exp.TOTAL_AGENT_TEST_SAMPLE_COUNT(), exp.TOTAL_ENV_STEP_TEST_SAMPLE_COUNT())
                self.assertEqual(exp.TOTAL_AGENT_TRAIN_SAMPLE_COUNT(), exp.TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT(), 500)

            return func

        single_exp_runner(wrap_algo(algo, locals), auto_choose_gpu_flag=False, gpu_id=0,
                          del_if_log_path_existed=True)

    def test_duplicate_exp(self):
        def func():
            GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT', dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=500,
                                                                    TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                                                    TOTAL_AGENT_UPDATE_COUNT=None))
            dqn, locals = self.create_dqn()
            env_spec = locals['env_spec']
            env = locals['env']
            agent = self.create_agent(env=locals['env'],
                                      algo=dqn,
                                      name='agent',
                                      eps=self.create_eps(env_spec)[0],
                                      env_spec=env_spec)[0]
            exp = self.create_exp(name='model_free', env=env, agent=agent)
            exp.run()

        base_path = GlobalConfig().DEFAULT_LOG_PATH
        duplicate_exp_runner(2, func, auto_choose_gpu_flag=False, gpu_id=0)
        self.assertTrue(os.path.isdir(base_path))
        self.assertTrue(os.path.isdir(os.path.join(base_path, 'exp_0')))
        self.assertTrue(os.path.isdir(os.path.join(base_path, 'exp_1')))
        self.assertTrue(os.path.isdir(os.path.join(base_path, 'exp_0', 'record')))
        self.assertTrue(os.path.isdir(os.path.join(base_path, 'exp_1', 'record')))
        self.assertTrue(os.path.isfile(os.path.join(base_path, 'exp_0', 'console.log')))
        self.assertTrue(os.path.isfile(os.path.join(base_path, 'exp_1', 'console.log')))

    def test_saving_scheduler_on_all_model_free_algo(self):
        to_test_algo_func = (self.create_ppo, self.create_dqn, self.create_ddpg)
        sample_traj_flag = (True, False, False)
        for i, func in enumerate(to_test_algo_func):
            self.setUp()
            single_exp_runner(_saving_scheduler(self, func,
                                                sample_traj_flag=sample_traj_flag[i]),
                              auto_choose_gpu_flag=False,
                              gpu_id=0,
                              del_if_log_path_existed=True)
            self.tearDown()

    def test_saving_scheduler_on_all_model_based_algo(self):
        to_test_algo_func = (self.create_mpc, self.create_dyna)
        for func in to_test_algo_func:
            self.setUp()
            single_exp_runner(_saving_scheduler(self, func), auto_choose_gpu_flag=False,
                              gpu_id=0, del_if_log_path_existed=True)
            self.tearDown()


def _saving_scheduler(self, creat_func=None, sample_traj_flag=False):
    def wrap_algo():
        def func(self, creat_func=None):
            GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT', dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=500,
                                                                    TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                                                    TOTAL_AGENT_UPDATE_COUNT=None))
            if not creat_func:
                algo, locals = self.create_dqn()
            else:
                algo, locals = creat_func()
            env_spec = locals['env_spec']
            env = locals['env']
            agent = self.create_agent(env=locals['env'],
                                      algo=algo,
                                      name='agent',
                                      eps=self.create_eps(env_spec)[0],
                                      env_spec=env_spec)[0]
            flow = None
            from baconian.algo.dyna import Dyna
            if isinstance(algo, Dyna):
                flow = self.create_dyna_flow(agent=agent, env=env)[0]
            exp = self.create_exp(name='model_free', env=env, agent=agent, flow=flow, traj_flag=sample_traj_flag)
            exp.run()
            self.assertEqual(exp.TOTAL_AGENT_TEST_SAMPLE_COUNT(), exp.TOTAL_ENV_STEP_TEST_SAMPLE_COUNT())
            self.assertEqual(exp.TOTAL_AGENT_TRAIN_SAMPLE_COUNT(), exp.TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT(), 500)

        return func(self, creat_func)

    return wrap_algo



from baconian.common.sampler.sample_data import TransitionData, TrajectoryData
from baconian.test.tests.set_up.setup import TestWithAll
from baconian.config.global_config import GlobalConfig


class TestPPO(TestWithAll):
    def test_init(self):
        ppo, locals = self.create_ppo()
        env = locals['env']
        env_spec = locals['env_spec']
        ppo.init()

        new_ppo, _ = self.create_ppo(name='new_ppo')
        new_ppo.copy_from(ppo)

        self.assert_var_list_id_no_equal(ppo.value_func.parameters('tf_var_list'),
                                         new_ppo.value_func.parameters('tf_var_list'))
        self.assert_var_list_id_no_equal(ppo.policy.parameters('tf_var_list'),
                                         new_ppo.policy.parameters('tf_var_list'))

        self.assert_var_list_equal(ppo.value_func.parameters('tf_var_list'),
                                   new_ppo.value_func.parameters('tf_var_list'))
        self.assert_var_list_equal(ppo.policy.parameters('tf_var_list'),
                                   new_ppo.policy.parameters('tf_var_list'))

        data = TransitionData(env_spec)
        st = env.reset()
        for i in range(100):
            ac = ppo.predict(st)
            assert ac.shape[0] == 1
            self.assertTrue(env_spec.action_space.contains(ac[0]))
            new_st, re, done, _ = env.step(ac)
            if i % 9 == 0 and i > 0:
                done = True
            else:
                done = False
            data.append(state=st, new_state=new_st, action=ac, reward=re, done=done)
        traj = TrajectoryData(env_spec=env_spec)
        traj.append(data)
        ppo.append_to_memory(traj)

        ppo.save(save_path=GlobalConfig().DEFAULT_LOG_PATH + '/ppo_test',
                 global_step=0,
                 name=ppo.name)
        for i in range(5):
            ppo.append_to_memory(traj)
            res = ppo.train()

            print('value_func_loss {}, policy_average_loss: {}'.format(res['value_func_loss'],
                                                                       res['policy_average_loss']))
            traj_data = TrajectoryData(env_spec=env_spec)
            traj_data.append(data)
            res = ppo.train(trajectory_data=traj_data,
                            train_iter=5,
                            sess=self.sess)

            print('value_func_loss {}, policy_average_loss: {}'.format(res['value_func_loss'],
                                                                       res['policy_average_loss']))

        self.assert_var_list_at_least_not_equal(ppo.value_func.parameters('tf_var_list'),
                                                new_ppo.value_func.parameters('tf_var_list'))
        self.assert_var_list_at_least_not_equal(ppo.policy.parameters('tf_var_list'),
                                                new_ppo.policy.parameters('tf_var_list'))

        ppo.load(path_to_model=GlobalConfig().DEFAULT_LOG_PATH + '/ppo_test',
                 model_name=ppo.name,
                 global_step=0)

        self.assert_var_list_equal(ppo.value_func.parameters('tf_var_list'),
                                   new_ppo.value_func.parameters('tf_var_list'))
        self.assert_var_list_equal(ppo.policy.parameters('tf_var_list'),
                                   new_ppo.policy.parameters('tf_var_list'))



from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
import numpy as np

OBS_DIM = 11
HID1_MULT = 10
HID3_SIZE = 5
HID1_SIZE = OBS_DIM * HID1_MULT
HID2_SIZE = int(np.sqrt(HID1_SIZE * HID3_SIZE))

POLICY_HID_MULTI = 10
ACT_DIM = 3
POLICY_HID3_SIZE = ACT_DIM * 10
POLICY_HID1_SIZE = OBS_DIM * POLICY_HID_MULTI
POLICY_HID2_SIZE = int(np.sqrt(POLICY_HID1_SIZE * POLICY_HID3_SIZE))

HOPPER_BENCHMARK_CONFIG_DICT = {
    'env_id': "Hopper-v2",
    'MLP_V': {
        'mlp_config': [
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "N_UNITS": HID1_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / OBS_DIM)
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "2",
                "N_UNITS": HID2_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / HID1_SIZE)
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "3",
                "N_UNITS": HID3_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / HID2_SIZE)
            },
            {
                "ACT": "IDENTITY",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "N_UNITS": 1,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / HID3_SIZE),
            }
        ]
    },
    'POLICY': {
        'mlp_config': [
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "N_UNITS": POLICY_HID1_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / OBS_DIM)
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "2",
                "N_UNITS": POLICY_HID2_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / POLICY_HID1_SIZE)
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "3",
                "N_UNITS": POLICY_HID3_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / POLICY_HID2_SIZE)
            },
            {
                "ACT": "IDENTITY",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "N_UNITS": ACT_DIM,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / POLICY_HID3_SIZE)
            }
        ]
    },
    'PPO': {
        'config_or_config_dict': {
            "gamma": 0.995,
            "lam": 0.98,
            "policy_train_iter": 20,
            "value_func_train_iter": 10,
            "clipping_range": None,
            "beta": 1.0,
            "eta": 50,
            "log_var_init": -1.0,
            "kl_target": 0.003,
            "policy_lr": 9e-4 / np.sqrt(POLICY_HID2_SIZE),
            "value_func_lr": 1e-2 / np.sqrt(HID2_SIZE),
            "value_func_train_batch_size": 256,
            "lr_multiplier": 1.0
        }
    },
    'TrainTestFlow': {
        "TEST_SAMPLES_COUNT": 5,
        "TRAIN_SAMPLES_COUNT": 20,
        'config_or_config_dict': {
            "TEST_EVERY_SAMPLE_COUNT": 5,
            "TRAIN_EVERY_SAMPLE_COUNT": 1,
            "START_TRAIN_AFTER_SAMPLE_COUNT": 1,
            "START_TEST_AFTER_SAMPLE_COUNT": 1000,
        }
    },
    'DEFAULT_EXPERIMENT_END_POINT': dict(TOTAL_AGENT_TRAIN_SAMPLE_FUNC_COUNT=1500,
                                         TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                         TOTAL_AGENT_UPDATE_COUNT=None),
}



from baconian.core.core import EnvSpec
from baconian.test.tests.set_up.setup import TestTensorflowSetup
from baconian.envs.gym_env import make


class TestDeterministicMLPPolicy(TestTensorflowSetup):
    def test_mlp_deterministic_policy(self):
        env = make('Pendulum-v0')
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)

        policy, locals = self.create_mlp_deterministic_policy(name='mlp_policy', env_spec=env_spec)
        policy.init()
        for _ in range(10):
            ac = policy.forward(obs=env.observation_space.sample())
            self.assertTrue(env.action_space.contains(ac[0]))
        p2 = policy.make_copy(name='test',
                              name_scope='test',
                              reuse=False)
        p2.init()
        self.assertGreater(len(policy.parameters('tf_var_list')), 0)
        self.assertGreater(len(p2.parameters('tf_var_list')), 0)
        for var1, var2 in zip(policy.parameters('tf_var_list'), p2.parameters('tf_var_list')):
            self.assertEqual(var1.shape, var2.shape)
            self.assertNotEqual(id(var1), id(var2))

        p3 = policy.make_copy(name='mlp_policy_2',
                              name_scope='mlp_policy',
                              reuse=True)
        p3.init()
        self.assertGreater(len(p3.parameters('tf_var_list')), 0)
        for var1, var2 in zip(policy.parameters('tf_var_list'), p3.parameters('tf_var_list')):
            self.assertEqual(var1.shape, var2.shape)
            self.assertEqual(id(var1), id(var2))


from baconian.core.flow.train_test_flow import Flow
from baconian.config.global_config import GlobalConfig
from baconian.common.logging import ConsoleLogger
from baconian.config.dict_config import DictConfig
from baconian.common.misc import *
from baconian.core.parameters import Parameters
from baconian.core.status import *


class MEPPO_Flow(Flow):
    """
    A typical flow for utilizing the model-based algo, it is not restricted to Dyna algorithms,
    but can be utilized by others.
    """

    required_func = ('train_algo', 'train_algo_from_synthesized_data', 'train_dynamics', 'test_algo', 'test_dynamics',
                     'sample_from_real_env', 'sample_from_dynamics_env', 'validate_policy_on_ensemble')

    required_key_dict = {
        "TEST_ALGO_EVERY_REAL_SAMPLE_COUNT": 1000,
        "TEST_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 1000,
        "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_REAL_ENV": 1000,
        "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_DYNAMICS_ENV": 1000,
        "TRAIN_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 1000,
        "START_TRAIN_ALGO_AFTER_SAMPLE_COUNT": 1,
        "START_TRAIN_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
        "START_TEST_ALGO_AFTER_SAMPLE_COUNT": 1,
        "START_TEST_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
        "WARM_UP_DYNAMICS_SAMPLES": 1000,
        "VALIDATION_THRESHOLD": 0.7,
        "VALIDATION_EVERY_FICTITIOUS_SET": 5,
        "SAMPLE_BEYOND_STOP_IMPROVEMENT": 10,
    }

    def __init__(self,
                 train_sample_count_func,
                 config_or_config_dict: (DictConfig, dict),
                 func_dict: dict, ):
        super().__init__(func_dict)
        super(MEPPO_Flow, self).__init__(func_dict=func_dict)
        config = construct_dict_config(config_or_config_dict, obj=self)
        self.parameters = Parameters(source_config=config, parameters=dict())
        self.time_step_func = train_sample_count_func
        self._last_train_algo_point = -1
        self._start_train_algo_point_from_dynamics = -1
        self._last_test_algo_point = -1
        self._start_train_dynamics_point = -1
        self._last_test_dynamics_point = -1
        self._last_performance = 0
        self._last_chance = 0
        self._fictitious_set_count = 0
        assert callable(train_sample_count_func)

    def _launch(self) -> bool:

        while True:

            if self._is_ended() is True:
                break

            real_batch_data = self._call_func('sample_from_real_env')

            if self.time_step_func() - self._start_train_dynamics_point <= self.parameters(
                        'TRAIN_DYNAMICS_EVERY_REAL_SAMPLE_COUNT') and \
                    self.time_step_func() > self.parameters('START_TRAIN_DYNAMICS_AFTER_SAMPLE_COUNT'):
                self._last_train_algo_point = self.time_step_func()
                self._call_func('train_dynamics', batch_data=real_batch_data)

            if self.time_step_func() <= self.parameters('WARM_UP_DYNAMICS_SAMPLES'):
                continue

            else:
                while True:
                    if self.time_step_func() - self._start_train_algo_point_from_dynamics <= self.parameters(
                                'TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_DYNAMICS_ENV') and \
                            self.time_step_func() > self.parameters(
                            'START_TRAIN_ALGO_AFTER_SAMPLE_COUNT'):
                        batch_data = self._call_func('sample_from_dynamics_env')
                        self._call_func('train_algo_from_synthesized_data', batch_data=batch_data)

                    else:
                        self._fictitious_set_count += 1
                        self._last_chance += 1
                        if self._fictitious_set_count >= self.parameters('VALIDATION_EVERY_FICTITIOUS_SET'):
                            performance = self._call_func('validate_policy_on_ensemble')

                            if performance < self.parameters('VALIDATION_THRESHOLD'):
                                if self._last_chance > self.parameters('SAMPLE_BEYOND_IMPROVEMENT'):
                                    self._last_chance = 0
                                    self._start_train_dynamics_point = self.time_step_func()
                                    break
                        self._start_train_algo_point_from_dynamics = self.time_step_func()
        return True

    def _is_ended(self):
        key_founded_flag = False
        finished_flag = False
        for key in GlobalConfig().DEFAULT_EXPERIMENT_END_POINT:
            if GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[key] is not None:
                key_founded_flag = True
                if get_global_status_collect()(key) >= GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[key]:
                    ConsoleLogger().print('info',
                                          'pipeline ended because {}: {} >= end point value {}'.
                                          format(key, get_global_status_collect()(key),
                                                 GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[key]))
                    finished_flag = True
        if key_founded_flag is False:
            ConsoleLogger().print(
                'warning',
                '{} in experiment_end_point is not registered with global status collector: {}, experiment may not end'.
                    format(GlobalConfig().DEFAULT_EXPERIMENT_END_POINT, list(get_global_status_collect()().keys())))
        return finished_flag


def create_meppo_flow(train_algo_func, train_algo_from_synthesized_data_func,
                      train_dynamics_func, test_algo_func, test_dynamics_func, sample_from_real_env_func,
                      sample_from_dynamics_env_func,
                      validate_policy_on_ensemble_func,
                      test_algo_every_real_sample_count,
                      test_dynamics_every_real_sample_count,
                      train_algo_every_real_sample_count_by_data_from_real_env,
                      train_algo_every_real_sample_count_by_data_from_dynamics_env,
                      train_dynamics_every_real_sample_count,
                      start_train_algo_after_sample_count,
                      start_train_dynamics_after_sample_count,
                      start_test_dynamics_after_sample_count,
                      start_test_algo_after_sample_count,
                      warm_up_dynamics_samples,
                      validation_threshold,
                      validation_every_fictitious_set,
                      sample_beyond_stop_improvement,
                      train_samples_counter_func=None):
    config_dict = dict(
        TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_REAL_ENV=train_algo_every_real_sample_count_by_data_from_real_env,
        TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_DYNAMICS_ENV=train_algo_every_real_sample_count_by_data_from_dynamics_env,
        TEST_ALGO_EVERY_REAL_SAMPLE_COUNT=test_algo_every_real_sample_count,
        TEST_DYNAMICS_EVERY_REAL_SAMPLE_COUNT=test_dynamics_every_real_sample_count,
        TRAIN_DYNAMICS_EVERY_REAL_SAMPLE_COUNT=train_dynamics_every_real_sample_count,
        START_TRAIN_ALGO_AFTER_SAMPLE_COUNT=start_train_algo_after_sample_count,
        START_TRAIN_DYNAMICS_AFTER_SAMPLE_COUNT=start_train_dynamics_after_sample_count,
        START_TEST_ALGO_AFTER_SAMPLE_COUNT=start_test_algo_after_sample_count,
        START_TEST_DYNAMICS_AFTER_SAMPLE_COUNT=start_test_dynamics_after_sample_count,
        WARM_UP_DYNAMICS_SAMPLES=warm_up_dynamics_samples,
        VALIDATION_THRESHOLD=validation_threshold,
        VALIDATION_EVERY_FICTITIOUS_SET=validation_every_fictitious_set,
        SAMPLE_BEYOND_STOP_IMPROVEMENT=sample_beyond_stop_improvement,
    )

    def return_func_dict(s_dict):
        return dict(func=s_dict[0],
                    args=s_dict[1],
                    kwargs=s_dict[2])

    func_dict = dict(
        train_algo=return_func_dict(train_algo_func),
        train_algo_from_synthesized_data=return_func_dict(train_algo_from_synthesized_data_func),
        train_dynamics=return_func_dict(train_dynamics_func),
        test_algo=return_func_dict(test_algo_func),
        test_dynamics=return_func_dict(test_dynamics_func),
        sample_from_real_env=return_func_dict(sample_from_real_env_func),
        sample_from_dynamics_env=return_func_dict(sample_from_dynamics_env_func),
        validate_policy_on_ensemble=return_func_dict(validate_policy_on_ensemble_func)
    )
    if train_samples_counter_func is None:
        def default_train_samples_counter_func():
            return get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT')

        train_samples_counter_func = default_train_samples_counter_func

    return MEPPO_Flow(config_or_config_dict=config_dict,
                      train_sample_count_func=train_samples_counter_func,
                      func_dict=func_dict)



from .policy import Policy, StochasticPolicy, DeterministicPolicy
from .constant_action_policy import ConstantActionPolicy
from .random_policy import UniformRandomPolicy
from .normal_distribution_mlp import NormalDistributionMLPPolicy
from .lqr_policy import LQRPolicy
from .ilqr_policy import iLQRPolicy
from .deterministic_mlp import DeterministicMLPPolicy



"""
For experiments, its functionality should include:
1. experiment and config set up
2. logging control
3. hyper-param tuning etc.
4. visualization
5. any related experiment utility
...
"""
from baconian.core.core import Basic
from baconian.common.logging import ConsoleLogger
from baconian.config.global_config import GlobalConfig
from baconian.core.tuner import Tuner
from baconian.core.util import init_func_arg_record_decorator
import tensorflow as tf
from typeguard import typechecked
from baconian.tf.util import create_new_tf_session
from baconian.core.core import Env
from baconian.core.agent import Agent
from baconian.common.logging import Recorder
from baconian.core.status import *
from baconian.core.flow.train_test_flow import Flow
from baconian.core.global_var import reset_all as reset_global_var
from baconian.common.logging import reset_logging


class Experiment(Basic):
    STATUS_LIST = ('CREATED', 'INITED', 'RUNNING', 'FINISHED', 'CORRUPTED')
    INIT_STATUS = 'CREATED'
    # required_key_dict = DictConfig.load_json(file_path=GlobalConfig().DEFAULT_EXPERIMENT_REQUIRED_KEY_LIST)
    required_key_dict = dict()

    @init_func_arg_record_decorator()
    @typechecked
    def __init__(self,
                 name: str,
                 agent: Agent,
                 env: Env,
                 flow: Flow,
                 tuner: Tuner = None,
                 register_default_global_status=True
                 ):

        """

        :param name: name of experiment
        :type name: str
        :param agent: agent of experiment
        :type agent: Agent
        :param env: environment of experiment
        :type env: Env
        :param flow: control flow to experiment
        :type flow: Flow
        :param tuner: hyper-parameter tuning method, currently in development
        :type tuner: Tuner
        :param register_default_global_status: register info key and status into global status collection
        :type register_default_global_status: bool
        """
        super().__init__(status=StatusWithSingleInfo(obj=self), name=name)
        self.agent = agent
        self.env = env
        self.tuner = tuner
        self.recorder = Recorder(flush_by_split_status=False, default_obj=self)
        # self.status_collector = StatusCollector()
        self.flow = flow
        if register_default_global_status is True:
            get_global_status_collect().register_info_key_status(obj=agent,
                                                                 info_key='predict_counter',
                                                                 under_status='TRAIN',
                                                                 return_name='TOTAL_AGENT_TRAIN_SAMPLE_COUNT')
            get_global_status_collect().register_info_key_status(obj=agent,
                                                                 info_key='sample_counter',
                                                                 under_status='TRAIN',
                                                                 return_name='TOTAL_AGENT_TRAIN_SAMPLE_FUNC_COUNT')

            get_global_status_collect().register_info_key_status(obj=agent,
                                                                 info_key='predict_counter',
                                                                 under_status='TEST',
                                                                 return_name='TOTAL_AGENT_TEST_SAMPLE_COUNT')
            get_global_status_collect().register_info_key_status(obj=agent,
                                                                 info_key='update_counter',
                                                                 under_status='TRAIN',
                                                                 return_name='TOTAL_AGENT_UPDATE_COUNT')
            get_global_status_collect().register_info_key_status(obj=env,
                                                                 info_key='step',
                                                                 under_status='TEST',
                                                                 return_name='TOTAL_ENV_STEP_TEST_SAMPLE_COUNT')
            get_global_status_collect().register_info_key_status(obj=env,
                                                                 info_key='step',
                                                                 under_status='TRAIN',
                                                                 return_name='TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT')

    def init(self):
        """ Create a new TensorFlow session, and set status to 'INITED'."""
        create_new_tf_session()
        self.agent.init()
        self.env.init()
        self.set_status('INITED')

    def run(self):
        """ Run the experiment, and set status to 'RUNNING'."""
        GlobalConfig().freeze()
        self.init()
        self.set_status('RUNNING')
        res = self.flow.launch()
        if res is False:
            self.set_status('CORRUPTED')
        else:
            self.set_status('FINISHED')
        self._exit()

    def _exit(self):
        """ Exit the experiment, reset global configurations and logging module."""
        sess = tf.get_default_session()
        if sess:
            sess.__exit__(None, None, None)
        tf.reset_default_graph()
        reset_global_status_collect()
        reset_logging()
        reset_global_var()
        GlobalConfig().unfreeze()

    def TOTAL_AGENT_UPDATE_COUNT(self):
        return get_global_status_collect()('TOTAL_AGENT_UPDATE_COUNT')

    def TOTAL_AGENT_TRAIN_SAMPLE_COUNT(self):
        return get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT')

    def TOTAL_AGENT_TEST_SAMPLE_COUNT(self):
        return get_global_status_collect()('TOTAL_AGENT_TEST_SAMPLE_COUNT')

    def TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT(self):
        return get_global_status_collect()('TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT')

    def TOTAL_ENV_STEP_TEST_SAMPLE_COUNT(self):
        return get_global_status_collect()('TOTAL_ENV_STEP_TEST_SAMPLE_COUNT')



from baconian.core.flow.train_test_flow import Flow
from baconian.config.global_config import GlobalConfig
from baconian.common.logging import ConsoleLogger
from baconian.config.dict_config import DictConfig
from baconian.common.misc import *
from baconian.core.parameters import Parameters
from baconian.core.status import *


class DynaFlow(Flow):
    """
    A typical flow for utilizing the model-based algo, it is not restricted to Dyna algorithms,
    but can be utilized by others.
    """

    required_func = ('train_algo', 'train_algo_from_synthesized_data', 'train_dynamics', 'test_algo', 'test_dynamics',
                     'sample_from_real_env', 'sample_from_dynamics_env')

    required_key_dict = {
        "TEST_ALGO_EVERY_REAL_SAMPLE_COUNT": 1000,
        "TEST_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 1000,
        "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_REAL_ENV": 1000,
        "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_DYNAMICS_ENV": 1000,
        "TRAIN_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 1000,
        "START_TRAIN_ALGO_AFTER_SAMPLE_COUNT": 1,
        "START_TRAIN_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
        "START_TEST_ALGO_AFTER_SAMPLE_COUNT": 1,
        "START_TEST_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
        "WARM_UP_DYNAMICS_SAMPLES": 1000
    }

    def __init__(self,
                 train_sample_count_func,
                 config_or_config_dict: (DictConfig, dict),
                 func_dict: dict, ):
        """

        :param train_sample_count_func: a function indicates how much training samples the agent has collected currently.
        :type train_sample_count_func: method
        :param config_or_config_dict: a Config or a dict should have the keys: (TEST_EVERY_SAMPLE_COUNT, TRAIN_EVERY_SAMPLE_COUNT, START_TRAIN_AFTER_SAMPLE_COUNT, START_TEST_AFTER_SAMPLE_COUNT)
        :type config_or_config_dict: Config or dict
        :param func_dict: function dict, holds the keys: 'sample', 'train', 'test'. each item in the dict as also should be a dict, holds the keys 'func', 'args', 'kwargs'
        :type func_dict: dict
        """
        super().__init__(func_dict)
        super(DynaFlow, self).__init__(func_dict=func_dict)
        config = construct_dict_config(config_or_config_dict, obj=self)
        self.parameters = Parameters(source_config=config, parameters=dict())
        self.time_step_func = train_sample_count_func
        self._last_train_algo_point = -1
        self._last_train_algo_point_from_dynamics = -1
        self._last_test_algo_point = -1
        self._last_train_dynamics_point = -1
        self._last_test_dynamics_point = -1
        assert callable(train_sample_count_func)

    def _launch(self) -> bool:

        """
        Launch the flow until it finished or catch a system-allowed errors
        (e.g., out of GPU memory, to ensure the log will be saved safely).

        :return: True if the flow correctly executed and finished
        :rtype: bool
        """
        while True:
            real_batch_data = self._call_func('sample_from_real_env')
            if self.time_step_func() - self.parameters(
                    'TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_REAL_ENV') >= self._last_train_algo_point and self.time_step_func() > self.parameters(
                'START_TRAIN_ALGO_AFTER_SAMPLE_COUNT'):
                self._last_train_algo_point = self.time_step_func()
                self._call_func('train_algo')
            if self.time_step_func() - self.parameters(
                    'TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_DYNAMICS_ENV') >= self._last_train_algo_point_from_dynamics and self.time_step_func() > self.parameters(
                'START_TRAIN_ALGO_AFTER_SAMPLE_COUNT') and self.time_step_func() >= self.parameters(
                'WARM_UP_DYNAMICS_SAMPLES'):
                batch_data = self._call_func('sample_from_dynamics_env')
                self._call_func('train_algo_from_synthesized_data', batch_data=batch_data)
                self._last_train_algo_point_from_dynamics = self.time_step_func()

            if self.time_step_func() - self.parameters(
                    'TRAIN_DYNAMICS_EVERY_REAL_SAMPLE_COUNT') >= self._last_train_dynamics_point and self.time_step_func() > self.parameters(
                'START_TRAIN_DYNAMICS_AFTER_SAMPLE_COUNT'):
                self._last_train_algo_point = self.time_step_func()
                self._call_func('train_dynamics', batch_data=real_batch_data)

            if self.time_step_func() - self.parameters(
                    'TEST_ALGO_EVERY_REAL_SAMPLE_COUNT') >= self._last_test_algo_point and self.time_step_func() > self.parameters(
                'START_TEST_ALGO_AFTER_SAMPLE_COUNT'):
                self._last_test_algo_point = self.time_step_func()
                self._call_func('test_algo')

            if self.time_step_func() - self.parameters(
                    'TEST_DYNAMICS_EVERY_REAL_SAMPLE_COUNT') >= self._last_test_dynamics_point and self.time_step_func() > self.parameters(
                'START_TEST_DYNAMICS_AFTER_SAMPLE_COUNT'):
                self._last_test_dynamics_point = self.time_step_func()
                self._call_func('test_dynamics')

            if self._is_ended() is True:
                break
        return True

    def _is_ended(self):
        """

        :return: True if an experiment is ended
        :rtype: bool
        """
        key_founded_flag = False
        finished_flag = False
        for key in GlobalConfig().DEFAULT_EXPERIMENT_END_POINT:
            if GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[key] is not None:
                key_founded_flag = True
                if get_global_status_collect()(key) >= GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[key]:
                    ConsoleLogger().print('info',
                                          'pipeline ended because {}: {} >= end point value {}'.
                                          format(key, get_global_status_collect()(key),
                                                 GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[key]))
                    finished_flag = True
        if key_founded_flag is False:
            ConsoleLogger().print(
                'warning',
                '{} in experiment_end_point is not registered with global status collector: {}, experiment may not end'.
                    format(GlobalConfig().DEFAULT_EXPERIMENT_END_POINT, list(get_global_status_collect()().keys())))
        return finished_flag


def create_dyna_flow(train_algo_func, train_algo_from_synthesized_data_func,
                     train_dynamics_func, test_algo_func, test_dynamics_func, sample_from_real_env_func,
                     sample_from_dynamics_env_func,
                     test_algo_every_real_sample_count,
                     test_dynamics_every_real_sample_count,
                     train_algo_every_real_sample_count_by_data_from_real_env,
                     train_algo_every_real_sample_count_by_data_from_dynamics_env,
                     train_dynamics_ever_real_sample_count,
                     start_train_algo_after_sample_count,
                     start_train_dynamics_after_sample_count,
                     start_test_dynamics_after_sample_count,
                     start_test_algo_after_sample_count,
                     warm_up_dynamics_samples,
                     train_samples_counter_func=None):
    config_dict = dict(
        TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_REAL_ENV=train_algo_every_real_sample_count_by_data_from_real_env,
        TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_DYNAMICS_ENV=train_algo_every_real_sample_count_by_data_from_dynamics_env,
        TEST_ALGO_EVERY_REAL_SAMPLE_COUNT=test_algo_every_real_sample_count,
        TEST_DYNAMICS_EVERY_REAL_SAMPLE_COUNT=test_dynamics_every_real_sample_count,
        TRAIN_DYNAMICS_EVERY_REAL_SAMPLE_COUNT=train_dynamics_ever_real_sample_count,
        START_TRAIN_ALGO_AFTER_SAMPLE_COUNT=start_train_algo_after_sample_count,
        START_TRAIN_DYNAMICS_AFTER_SAMPLE_COUNT=start_train_dynamics_after_sample_count,
        START_TEST_ALGO_AFTER_SAMPLE_COUNT=start_test_algo_after_sample_count,
        START_TEST_DYNAMICS_AFTER_SAMPLE_COUNT=start_test_dynamics_after_sample_count,
        WARM_UP_DYNAMICS_SAMPLES=warm_up_dynamics_samples,
    )

    def return_func_dict(s_dict):
        return dict(func=s_dict[0],
                    args=s_dict[1],
                    kwargs=s_dict[2])

    func_dict = dict(
        train_algo=return_func_dict(train_algo_func),
        train_algo_from_synthesized_data=return_func_dict(train_algo_from_synthesized_data_func),
        train_dynamics=return_func_dict(train_dynamics_func),
        test_algo=return_func_dict(test_algo_func),
        test_dynamics=return_func_dict(test_dynamics_func),
        sample_from_real_env=return_func_dict(sample_from_real_env_func),
        sample_from_dynamics_env=return_func_dict(sample_from_dynamics_env_func),
    )
    if train_samples_counter_func is None:
        def default_train_samples_counter_func():
            return get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT')

        train_samples_counter_func = default_train_samples_counter_func

    return DynaFlow(config_or_config_dict=config_dict,
                    train_sample_count_func=train_samples_counter_func,
                    func_dict=func_dict)



from typeguard import typechecked
from baconian.tf.util import MLPCreator
import tensorflow as tf
import numpy as np
from baconian.tf.tf_parameters import ParametersWithTensorflowVariable


class MLP(object):

    @typechecked
    def __init__(self,
                 input_ph: tf.Tensor,
                 name_scope: str,
                 net_name: str,
                 reuse,
                 mlp_config: list,
                 input_norm: np.ndarray = None,
                 output_norm: np.ndarray = None,
                 output_low: np.ndarray = None,
                 output_high: np.ndarray = None,
                 ):
        self.input_ph = input_ph
        self.name_scope = name_scope
        self.mlp_config = mlp_config
        self.mlp_net_name = net_name
        self.net, self.output, self.var_list = MLPCreator.create_network_with_tf_layers(input=input_ph,
                                                                                        reuse=reuse,
                                                                                        network_config=mlp_config,
                                                                                        tf_var_scope=name_scope,
                                                                                        net_name=net_name,
                                                                                        input_norm=input_norm,
                                                                                        output_high=output_high,
                                                                                        output_low=output_low,
                                                                                        output_norm=output_norm)
        for var in self.var_list:
            assert name_scope in var.name
        self._parameters = ParametersWithTensorflowVariable(tf_var_list=self.var_list,
                                                            name='parameters_{}'.format(self.mlp_net_name),
                                                            rest_parameters=dict())

    def forward(self, input: np.ndarray, sess=tf.get_default_session()) -> np.ndarray:
        feed_dict = {
            self.input_ph: input,
            **self._parameters.return_tf_parameter_feed_dict()
        }
        res = sess.run(self.output,
                       feed_dict=feed_dict)
        return np.squeeze(res)

    def copy_from(self, obj) -> bool:
        if not isinstance(obj, type(self)):
            raise TypeError('Wrong type of obj %s to be copied, which should be %s' % (type(obj), type(self)))
        self._parameters.copy_from(source_parameter=obj._parameters)
        return True

    def init(self, source_obj=None):
        self._parameters.init()
        if source_obj:
            self.copy_from(obj=source_obj)



from baconian.test.tests.set_up.setup import TestWithAll
from baconian.algo.dynamics.reward_func.reward_func import CostFunc
from baconian.envs.gym_env import make
import numpy as np
from baconian.core.core import EnvSpec
from baconian.algo.dynamics.dynamics_model import GlobalDynamicsModel
from baconian.algo.policy.ilqr_policy import iLQRPolicy
from baconian.algo.dynamics.dynamics_model import DynamicsEnvWrapper
from baconian.algo.dynamics.terminal_func.terminal_func import RandomTerminalFunc


class DebugDynamics(GlobalDynamicsModel):
    flag = 0.5
    st = None

    def _state_transit(self, state, action, **kwargs) -> np.ndarray:
        return state + 0.0001 * action
        # self.flag *= -1.0
        # return np.ones_like(self.env_spec.obs_space.sample()) * self.flag
        # return self.env_spec.obs_space.sample()


class DebuggingCostFunc(CostFunc):
    def __call__(self, state=None, action=None, new_state=None, **kwargs) -> float:
        # return float(np.sum(action * action) + np.sum(state * state))
        return float(np.sum(action + action * action))


class TestiLQRPolicy(TestWithAll):
    def test_correctness(self):
        env_id = 'Pendulum-v0'

        env = make(env_id)
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        dyna = DebugDynamics(env_spec=env_spec)
        dyna = DynamicsEnvWrapper(dynamics=dyna)
        dyna.set_terminal_reward_func(terminal_func=RandomTerminalFunc(),
                                      reward_func=DebuggingCostFunc())
        policy = iLQRPolicy(env_spec=env_spec,
                            T=10,
                            delta=0.05,
                            iteration=2,
                            dynamics=dyna,
                            dynamics_model_train_iter=10,
                            cost_fn=DebuggingCostFunc())
        st = env.reset()
        dyna.st = np.zeros_like(st)
        for i in range(10):
            ac = policy.forward(st)
            st, _, _, _ = env.step(st)
            # st = dyna.step(action=ac, state=st)
            print("analytical optimal action -0.5, cost -0.25")
            print('state: {}, action: {}, cost {}'.format(st, ac, policy.iLqr_instance.cost_fn(state=st, action=ac,
                                                                                               new_state=None)))



from baconian.test.tests.set_up.setup import BaseTestCase
from baconian.common.noise import *
from baconian.common.schedules import *

t = 0


def get_t():
    global t
    return t


class TestNoise(BaseTestCase):
    def test_all_noise(self):
        action_w = LinearScheduler(t_fn=get_t,
                                   schedule_timesteps=100,
                                   final_p=1.0,
                                   initial_p=0.0)
        noise_w = LinearScheduler(t_fn=get_t,
                                  final_p=0.0,
                                  schedule_timesteps=100,
                                  initial_p=1.0)

        noise_wrapper = AgentActionNoiseWrapper(noise=OUNoise(),
                                                action_weight_scheduler=action_w,
                                                noise_weight_scheduler=noise_w)
        for i in range(101):
            print('action w {}, noise w {}'.format(noise_wrapper.action_weight_scheduler.value(),
                                                   noise_wrapper.noise_weight_scheduler.value()))
            print(noise_wrapper(action=1.0))
            if i == 100:
                self.assertEqual(noise_wrapper(action=1.0), 1.0)
            global t
            t += 1



import tensorflow as tf
from baconian.core.parameters import Parameters
from baconian.config.global_config import GlobalConfig
from overrides.overrides import overrides
from typeguard import typechecked
import os
from baconian.common.schedules import Scheduler
import numpy as np


class ParametersWithTensorflowVariable(Parameters):

    @typechecked
    def __init__(self, tf_var_list: list, rest_parameters: dict, name: str,
                 max_to_keep=GlobalConfig().DEFAULT_MAX_TF_SAVER_KEEP,
                 default_save_type='tf',
                 source_config=None,
                 to_scheduler_param_tuple: list = None,
                 save_rest_param_flag=True,
                 to_ph_parameter_dict: dict = None,
                 require_snapshot=False):
        super(ParametersWithTensorflowVariable, self).__init__(parameters=rest_parameters,
                                                               name=name,
                                                               to_scheduler_param_tuple=to_scheduler_param_tuple,
                                                               source_config=source_config)
        self._tf_var_list = tf_var_list
        self.snapshot_var = []
        self.save_snapshot_op = []
        self.load_snapshot_op = []
        self.saver = None
        self.max_to_keep = max_to_keep
        self.require_snapshot = require_snapshot
        self.default_checkpoint_type = default_save_type
        self.save_rest_param_flag = save_rest_param_flag
        if default_save_type != 'tf':
            raise NotImplementedError('only support saving tf')
        self._registered_tf_ph_dict = dict()
        if to_ph_parameter_dict:
            for key, val in to_ph_parameter_dict.items():
                self.to_tf_ph(key=key, ph=val)

    @overrides
    def init(self):
        Parameters.init(self)
        sess = tf.get_default_session()
        sess.run(tf.variables_initializer(var_list=self._tf_var_list))
        if self.require_snapshot is True:
            if len(self.snapshot_var) == 0:
                # add the snapshot op after the init
                sess = tf.get_default_session()
                with tf.variable_scope('snapshot'):
                    for var in self._tf_var_list:
                        snap_var = tf.Variable(initial_value=sess.run(var),
                                               expected_shape=var.get_shape().as_list(),
                                               name=str(var.name).split(':')[0])
                        self.snapshot_var.append(snap_var)
                        self.save_snapshot_op.append(tf.assign(snap_var, var))
                        self.load_snapshot_op.append(tf.assign(var, snap_var))
            sess.run(tf.variables_initializer(var_list=self.snapshot_var))
            sess.run(self.save_snapshot_op)
        self.saver = tf.train.Saver(max_to_keep=self.max_to_keep,
                                    var_list=self._tf_var_list)

    def return_tf_parameter_feed_dict(self) -> dict:
        res = dict()
        for key, val in self._registered_tf_ph_dict.items():
            res[val] = self(key, require_true_value=True)
        return res

    def save_snapshot(self):
        sess = tf.get_default_session()
        if len(self.save_snapshot_op) == 0:
            with tf.variable_scope('snapshot'):
                for var in self._tf_var_list:
                    snap_var = tf.Variable(initial_value=sess.run(var),
                                           expected_shape=var.get_shape().as_list(),
                                           name=var.name)
                    self.snapshot_var.append(snap_var)
                    self.save_snapshot_op.append(tf.assign(var, snap_var))
        sess.run(self.save_snapshot_op)

    def load_snapshot(self):
        sess = tf.get_default_session()
        if len(self.load_snapshot_op) == 0:
            with tf.variable_scope('snapshot'):
                for var in self._tf_var_list:
                    snap_var = tf.Variable(initial_value=sess.run(var),
                                           expected_shape=var.get_shape().as_list(),
                                           name=var.name)
                    self.snapshot_var.append(snap_var)
                    self.load_snapshot_op.append(tf.assign(snap_var, var))
        sess.run(self.load_snapshot_op)

    def save(self, save_path, global_step, sess=None, name=None, *args, **kwargs):
        if self.default_checkpoint_type == 'tf':
            self._save_to_tf(save_path=save_path,
                             global_step=global_step,
                             sess=sess,
                             name=name)
        elif self.default_checkpoint_type == 'h5py':
            raise NotImplementedError
        if self.save_rest_param_flag is False:
            to_save_dict = dict(_source_config=self._source_config.config_dict)
        else:
            to_save_dict = dict(_parameters=self._parameters, _source_config=self._source_config.config_dict)
        Parameters.save(self,
                        save_path=save_path,
                        global_step=global_step,
                        default_save_param=to_save_dict,
                        name=name)

    def load(self, path_to_model, global_step=None, sess=None, model_name=None, *args, **kwargs):
        if not model_name:
            model_name = self.name
        if self.default_checkpoint_type == 'tf':
            self._load_from_tf(path_to_model=path_to_model,
                               global_step=global_step,
                               sess=sess, model_name=model_name)
        elif self.default_checkpoint_type == 'h5py':
            self._load_from_h5py(*args, **kwargs)
        Parameters.load(self,
                        load_path=path_to_model,
                        global_step=global_step,
                        name=model_name)

    def _save_to_tf(self, save_path, global_step, sess=None, name=None):
        name = name if name else self.name
        sess = sess if sess else tf.get_default_session()
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        save_path = os.path.join(save_path, name)
        self.saver.save(sess=sess,
                        save_path=save_path,
                        global_step=global_step)

    def _load_from_tf(self, path_to_model, model_name, global_step=None, sess=None):
        sess = sess if sess else tf.get_default_session()

        if not global_step:
            loaded_path = tf.train.latest_checkpoint(path_to_model)

        else:
            loaded_path = os.path.join(os.path.join(path_to_model, '{}-{}'.format(model_name, global_step)))
        self.saver.restore(sess=sess,
                           save_path=loaded_path)

    def _save_to_h5py(self, var_list, sess):
        raise NotImplementedError

    def _load_from_h5py(self, path_to_h5py):
        raise NotImplementedError

    def __call__(self, key=None, require_true_value=False):
        if key in self._registered_tf_ph_dict.keys():
            if require_true_value is True:
                return super().__call__(key)
            else:
                return self._registered_tf_ph_dict[key]
        else:
            if key == 'tf_var_list':
                return self._tf_var_list
            else:
                return super().__call__(key)

    def set(self, key, new_val):
        if not isinstance(new_val, type(self(key, require_true_value=True))):
            raise TypeError('new value of parameters {} should be type {} instead of {}'.format(key, type(self(key)),
                                                                                                type(new_val)))
        else:
            if key == 'tf_var_list':
                self.set_tf_var_list(new_val)
            elif key in self._parameters:
                self._parameters[key] = new_val
            else:
                self._source_config.set(key, new_val)

    def set_tf_var_list(self, tf_var_list: list):
        temp_var_list = list(set(tf_var_list))
        if len(temp_var_list) < len(tf_var_list):
            raise ValueError('Redundant tf variable in tf_var_list')
        for var in tf_var_list:
            assert isinstance(var, (tf.Tensor, tf.Variable))
        self._tf_var_list += tf_var_list

    def to_tf_ph(self, key, ph: tf.Tensor):
        # call the parameters first to make sure it have an init value
        self(key)
        self._registered_tf_ph_dict[key] = ph

    @overrides
    def copy_from(self, source_parameter, deep_copy=None):
        if not isinstance(source_parameter, type(self)):
            raise TypeError()
        super(ParametersWithTensorflowVariable, self).copy_from(source_parameter)
        tmp_op_list = []
        for t_para, s_para in zip(self._tf_var_list, source_parameter._tf_var_list):
            tmp_op_list.append(tf.assign(t_para, s_para))
        sess = tf.get_default_session()
        sess.run(tmp_op_list)
        del tmp_op_list

    def _update_dict(self, source_dict: dict, target_dict: dict):
        for key, val in source_dict.items():
            if isinstance(val, tf.Tensor):
                continue
            target_dict[key] = val

    @typechecked
    def set_scheduler(self, param_key: str, scheduler: Scheduler, to_tf_ph_flag=True):
        ori_val = self(param_key)
        if to_tf_ph_flag is True:
            self.to_tf_ph(key=param_key,
                          ph=tf.placeholder(shape=tuple(np.array(ori_val).shape),
                                            dtype=tf.dtypes.as_dtype(np.array(ori_val).dtype)))
        scheduler.initial_p = ori_val
        self._scheduler_info_dict[param_key] = dict(param_key=param_key, scheduler=scheduler)




MOUNTAINCAR_BENCHMARK_CONFIG_DICT = {
    'env_id': "MountainCar-v0",
    'MLPQValueFunction': {
        'mlp_config': [
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "N_UNITS": 64,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "2",
                "N_UNITS": 64,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "RELU",
                "B_INIT_VALUE": 0.0,
                "NAME": "3",
                "N_UNITS": 256,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "LINEAR",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "N_UNITS": 1,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            }
        ]
    },
    'DQN': {
        'config_or_config_dict': {
            "REPLAY_BUFFER_SIZE": 50000,
            "GAMMA": 0.99,
            "DECAY": 0,
            "BATCH_SIZE": 32,
            "TRAIN_ITERATION": 1,
            "LEARNING_RATE": 0.001,
        },
        'replay_buffer': None
    },
    'EpsilonGreedy': {
        'LinearScheduler': {
            'initial_p': 1.0,
            'final_p': 0.1,
            'schedule_timesteps': int(0.1 * 100000)
        },
        'config_or_config_dict': {
            "init_random_prob": 1.0
        }
    },
    'TrainTestFlow': {
        "TEST_SAMPLES_COUNT": 3,
        "TRAIN_SAMPLES_COUNT": 1,
        'config_or_config_dict': {
            "TEST_EVERY_SAMPLE_COUNT": 100,
            "TRAIN_EVERY_SAMPLE_COUNT": 1,
            "START_TRAIN_AFTER_SAMPLE_COUNT": 1000,
            "START_TEST_AFTER_SAMPLE_COUNT": 0,
        }
    },

    'DEFAULT_EXPERIMENT_END_POINT': dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=100000,
                                         TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                         TOTAL_AGENT_UPDATE_COUNT=None),
}



from baconian.test.tests.set_up.setup import TestWithAll
import numpy as np
from baconian.algo.dynamics.linear_dynamics_model import LinearDynamicsModel, LinearRegressionDynamicsModel
from baconian.common.data_pre_processing import RunningStandardScaler


class TestDynamicsModel(TestWithAll):

    def test_dynamics_model(self):
        real_env = self.create_env('Pendulum-v0')
        x = real_env.observation_space.flat_dim
        u = real_env.action_space.flat_dim
        a = LinearDynamicsModel(env_spec=real_env.env_spec,
                                state_transition_matrix=np.ones((x,
                                                                 x + u)) * 0.01,
                                bias=np.ones(x) * 0.02)
        new_state = a.step(action=np.ones_like(real_env.action_space.sample()),
                           state=np.ones_like(real_env.observation_space.sample()))
        print('new state', new_state)
        true_new = np.ones([x]) * (x + u) * 0.01 + np.ones([x]) * 0.02
        print('true state', true_new)
        self.assertTrue(np.equal(true_new, new_state).all())

    def test_linear_regression_model(self):
        real_env = self.create_env('Pendulum-v0')
        real_env.init()
        x = real_env.observation_space.flat_dim
        u = real_env.action_space.flat_dim
        a = LinearRegressionDynamicsModel(env_spec=real_env.env_spec,
                                          state_input_scaler=RunningStandardScaler(
                                              dims=real_env.observation_space.flat_dim),
                                          action_input_scaler=RunningStandardScaler(
                                              dims=real_env.action_space.flat_dim),
                                          state_output_scaler=RunningStandardScaler(
                                              dims=real_env.observation_space.flat_dim))
        data = self.sample_transition(env=real_env, count=100)
        a.train(batch_data=data)
        gen = data.return_generator(shuffle_flag=True)
        predict = []
        for state, _, action, _, _ in gen:
            predict.append(a.step(state=state, action=action))
        print(np.linalg.norm(np.array(predict) - data.new_state_set, ord=1))
        print(np.linalg.norm(np.array(predict) - data.new_state_set, ord=2))



from baconian.algo.dqn import DQN
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.core.agent import Agent
from baconian.algo.misc import EpsilonGreedy
from baconian.core.experiment import Experiment
from baconian.core.flow.train_test_flow import create_train_test_flow
from baconian.config.global_config import GlobalConfig
from baconian.common.schedules import LinearScheduler
from baconian.core.status import get_global_status_collect


def task_fn():
    env = make('Acrobot-v1')
    name = 'demo_exp'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)
    mlp_q = MLPQValueFunction(env_spec=env_spec,
                              name_scope=name + '_mlp_q',
                              name=name + '_mlp_q',
                              mlp_config=[
                                  {
                                      "ACT": "TANH",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "1",
                                      "N_UNITS": 64,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "TANH",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "2",
                                      "N_UNITS": 64,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "RELU",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "3",
                                      "N_UNITS": 256,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "LINEAR",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "OUPTUT",
                                      "N_UNITS": 1,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  }
                              ])
    dqn = DQN(env_spec=env_spec,
              config_or_config_dict=dict(REPLAY_BUFFER_SIZE=50000,
                                         GAMMA=0.99,
                                         BATCH_SIZE=32,
                                         LEARNING_RATE=0.001,
                                         TRAIN_ITERATION=1,
                                         DECAY=0),
              name=name + '_dqn',
              value_func=mlp_q)
    agent = Agent(env=env, env_spec=env_spec,
                  algo=dqn,
                  name=name + '_agent',
                  exploration_strategy=EpsilonGreedy(action_space=env_spec.action_space,
                                                     prob_scheduler=LinearScheduler(
                                                         t_fn=lambda: get_global_status_collect()(
                                                             'TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                                                         schedule_timesteps=int(0.1 * 100000),
                                                         initial_p=1.0,
                                                         final_p=0.02),
                                                     init_random_prob=0.1),
                  noise_adder=None)

    flow = create_train_test_flow(
        test_every_sample_count=100,
        train_every_sample_count=1,
        start_test_after_sample_count=0,
        start_train_after_sample_count=1000,
        sample_func_and_args=(agent.sample, (), dict(sample_count=1,
                                                     env=agent.env,
                                                     store_flag=True)),
        train_func_and_args=(agent.train, (), dict()),
        test_func_and_args=(agent.test, (), dict(sample_count=3)),
    )
    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name
    )
    experiment.run()


from baconian.core.experiment_runner import *

GlobalConfig().set('DEFAULT_LOG_PATH', './log_path')
single_exp_runner(task_fn, del_if_log_path_existed=True)



from baconian.core.core import Env, EnvSpec

have_mujoco_flag = True
try:
    from dm_control import mujoco
    from gym.envs.mujoco import mujoco_env
    from dm_control import suite
    from dm_control.rl.specs import ArraySpec
    from dm_control.rl.specs import BoundedArraySpec
    from collections import OrderedDict
except Exception:
    have_mujoco_flag = False

import numpy as np
import types
from gym.spaces import *
import baconian.common.spaces as garage_space


def convert_dm_control_to_gym_space(dm_control_space):
    r"""Convert dm_control space to gym space. """
    if isinstance(dm_control_space, BoundedArraySpec):
        space = Box(low=dm_control_space.minimum,
                    high=dm_control_space.maximum,
                    dtype=dm_control_space.dtype)
        assert space.shape == dm_control_space.shape
        return garage_space.Box(low=space.low, high=space.high)
    elif isinstance(dm_control_space, ArraySpec) and not isinstance(dm_control_space, BoundedArraySpec):
        space = Box(low=-float('inf'),
                    high=float('inf'),
                    shape=dm_control_space.shape,
                    dtype=dm_control_space.dtype)
        return garage_space.Box(low=space.low, high=space.high)
    elif isinstance(dm_control_space, OrderedDict):
        space = Dict(OrderedDict([(key, convert_dm_control_to_gym_space(value))
                                  for key, value in dm_control_space.items()]))
        return garage_space.Dict(space.spaces)
    else:
        raise NotImplementedError


_env_inited_count = dict()


# def make(gym_env_id, allow_multiple_env=True):
#     """
#
#     :param gym_env_id:
#     :param allow_multiple_env:
#     :return:
#     """
#     if allow_multiple_env is True:
#
#         if gym_env_id not in _env_inited_count:
#             _env_inited_count[gym_env_id] = 0
#         else:
#             _env_inited_count[gym_env_id] += 1
#
#         return GymEnv(gym_env_id, name='{}_{}'.format(gym_env_id, _env_inited_count[gym_env_id]))
#     else:
#         return GymEnv(gym_env_id)


class DMControlEnv(Env):
    """
    DeepMind Control Suite environment wrapping module
    """

    def __init__(self, dmcs_env_id: str, name: str = None):
        """

        :param dmcs_env_id:
        :param name:
        """
        super().__init__(name=name if name else dmcs_env_id)
        self.env_id = dmcs_env_id
        self.timestep = {}
        try:
            self.env = suite.load(dmcs_env_id, name)
        except ValueError:
            raise ValueError('Env id: {} and task: {} is not supported currently'.format(dmcs_env_id, name))

        self.metadata = {'render.modes': ['human', 'rgb_array'],
                         'video.frames_per_second': int(np.round(1.0 / self.env.control_timestep()))}

        self.action_space = convert_dm_control_to_gym_space(self.env.action_spec())
        self.observation_space = convert_dm_control_to_gym_space(self.env.observation_spec())
        if isinstance(self.action_space, garage_space.Box):
            self.action_space.low = np.nan_to_num(self.action_space.low)
            self.action_space.high = np.nan_to_num(self.action_space.high)
            self.action_space.sample = types.MethodType(self._sample_with_nan, self.action_space)
        if isinstance(self.observation_space, garage_space.Box):
            self.observation_space.low = np.nan_to_num(self.observation_space.low)
            self.observation_space.high = np.nan_to_num(self.observation_space.high)
            self.observation_space.sample = types.MethodType(self._sample_with_nan, self.observation_space)

        self.env_spec = EnvSpec(obs_space=self.observation_space,
                                action_space=self.action_space)

        self.viewer = None

    def step(self, action):
        """

        :param action:
        :return:
        """
        super().step(action)
        self.timestep = self.env.step(action)
        observation = self.timestep.observation
        reward = self.timestep.reward
        done = self.timestep.last()
        info = {}
        return observation, reward, done, info

    def reset(self):
        """

        :return:
        """
        super().reset()

        return self.env.reset()

    def init(self):
        """

        :return:
        """
        super().init()

        return self.reset()

    def seed(self, seed=None):
        """

        :param seed:
        :return:
        """
        return self.env.task.random.seed(seed)

    def get_state(self):
        """

        :return:
        """
        if self.timestep != {}:
            return self.timestep.observation
        else:
            raise ValueError('Env id: {} does not have an observation yet.'.format(self.env_id))

    @staticmethod
    def _sample_with_nan(space: garage_space.Space):
        """

        :param space:
        :return:
        """
        assert isinstance(space, garage_space.Box)
        high = np.ones_like(space.low)
        low = -1 * np.ones_like(space.high)
        return np.clip(np.random.uniform(low=low, high=high, size=space.low.shape),
                       a_min=space.low,
                       a_max=space.high)


if __name__ == '__main__':
    a = suite.load("cartpole", "swingup")
    a = DMControlEnv("cartpole", "swingup")



from baconian.core.core import Env
from baconian.common.spaces import Box
from baconian.envs.gym_env import GymEnv
import numpy as np


class Wrapper(Env):

    def __init__(self, env: Env):
        if isinstance(env, GymEnv):
            self.env = env.unwrapped_gym
            self.src_env = env
        else:
            self.env = env
            self.src_env = env
        super().__init__(name=env.name + '_wrapper', copy_from_env=env)

    def __getattr__(self, item):
        if hasattr(self.src_env, item):
            return getattr(self.src_env, item)
        if hasattr(self.env, item):
            return getattr(self.env, item)
        raise AttributeError()

    def seed(self, seed=None):
        return self.src_env.seed(seed)

    @property
    def unwrapped(self):
        return self.env.unwrapped

    @property
    def spec(self):
        return self.env.spec

    @classmethod
    def class_name(cls):
        return cls.__name__

    def __str__(self):
        return '<{}{}>'.format(type(self).__name__, self.env)

    def __repr__(self):
        return str(self)

    def reset(self):
        return self.src_env.reset()

    def get_state(self):
        return self.src_env.get_state()


class ObservationWrapper(Wrapper):
    def reset(self):
        observation = self.src_env.reset()
        return self._observation(observation)

    def step(self, action):
        observation, reward, done, info = self.src_env.step(action)
        return self.observation(observation), reward, done, info

    def observation(self, observation):
        return self._observation(observation)

    def _observation(self, observation):
        raise NotImplementedError

    def get_state(self):
        return self._observation(self.src_env.get_state())


class RewardWrapper(Wrapper):
    def step(self, action):
        observation, reward, done, info = self.src_env.step(action)
        return observation, self.reward(observation, action, reward, done, info), done, info

    def reward(self, observation, action, reward, done, info):
        return self._reward(observation, action, reward, done, info)

    def _reward(self, observation, action, reward, done, info):
        raise NotImplementedError


class ActionWrapper(Wrapper):
    def step(self, action):
        action = self.action(action)
        return self.src_env.step(action)

    def action(self, action):
        return self._action(action)

    def _action(self, action):
        raise NotImplementedError

    def reverse_action(self, action):
        return self._reverse_action(action)

    def _reverse_action(self, action):
        raise NotImplementedError


class StepObservationWrapper(ObservationWrapper):
    def __init__(self, env: Env, step_limit=100000):
        super().__init__(env=env)
        assert isinstance(self.src_env.observation_space, Box), 'not support non Box space for step observation wrapper'
        self.src_env.observation_space = Box(low=np.concatenate([self.src_env.observation_space.low, np.array([0])]),
                                             high=np.concatenate(
                                                 [self.src_env.observation_space.high, np.array([step_limit])]))
        self.src_env.env_spec.obs_space = self.src_env.observation_space
        self.observation_space = self.src_env.observation_space

    def _observation(self, observation):
        obs = np.array(observation)
        return np.concatenate([obs, np.array([self.src_env.trajectory_level_step_count])])



from baconian.test.tests.set_up.setup import TestWithAll
import numpy as np
from baconian.algo.dynamics.gaussian_process_dynamiocs_model import GaussianProcessDyanmicsModel
from baconian.common.sampler.sample_data import TransitionData
from baconian.core.core import EnvSpec
import pandas as pd
 

def get_some_samples(env, num, env_spec, policy):
    data = TransitionData(env_spec=env_spec)
    st = env.reset()
    for i in range(num):
        ac = policy.forward(st)
        new_st, re, _, _ = env.step(ac)
        data.append(state=st, new_state=new_st, action=ac, reward=re, done=False)
        st = new_st
    return data


class TestDynamicsModel(TestWithAll):

    # def test_more(self):
    #     for i in range(10):
    #         var = self.test_dynamics_model_in_pendulum()
    #         for v in var:
    #             del v

    def test_dynamics_model_in_pendulum(self):
        env = self.create_env('Pendulum-v0')
        env_spec = EnvSpec(obs_space=env.observation_space, action_space=env.action_space)
        policy, _ = self.create_uniform_policy(env_spec=env_spec)
        policy.allow_duplicate_name = True
        data = get_some_samples(env=env, policy=policy, num=100, env_spec=env_spec)
        gp = GaussianProcessDyanmicsModel(env_spec=env_spec, batch_data=data)
        gp.allow_duplicate_name = True
        gp.init()
        gp.train()
        print("gp first fit")
        for i in range(len(data.state_set)):
            res = gp.step(action=data.action_set[i],
                          state=data.state_set[i],
                          allow_clip=True)
            _, var = gp._state_transit(action=data.action_set[i],
                                       state=data.state_set[i],
                                       required_var=True)
            print(res)
            print(data.new_state_set[i])
            print(np.sqrt(var))
            try:
                self.assertTrue(np.isclose(res,
                                           data.new_state_set[i], atol=1e-2).all())
                self.assertTrue(np.greater_equal(data.new_state_set[i], res - 10.0 * np.sqrt(var)).all())
                self.assertTrue(np.less_equal(data.new_state_set[i], res + 10.0 * np.sqrt(var)).all())
            except Exception as e:
                print(e)
                print(np.isclose(res, data.new_state_set[i], atol=1e-2).all())
                print(np.greater_equal(data.new_state_set[i], res - 10.0 * np.sqrt(var)).all())
                print(np.less_equal(data.new_state_set[i], res + 10.0 * np.sqrt(var)).all())
                raise e

        lengthscales = {}
        variances = {}
        noises = {}
        for i, model in enumerate(gp.mgpr_model.models):
            lengthscales['GP' + str(i)] = model.kern.lengthscales.value
            variances['GP' + str(i)] = np.array([model.kern.variance.value])
            noises['GP' + str(i)] = np.array([model.likelihood.variance.value])
        print('-----Learned models------')
        pd.set_option('precision', 3)
        print('---Lengthscales---')
        print(pd.DataFrame(data=lengthscales))
        print('---Variances---')
        print(pd.DataFrame(data=variances))
        print('---Noises---')
        print(pd.DataFrame(data=noises))

        # re fit the gp
        print("gp re fit")

        data = get_some_samples(env=env, policy=policy, num=100, env_spec=env_spec)
        gp.train(batch_data=data)
        for i in range(len(data.state_set)):
            res = gp.step(action=data.action_set[i],
                          state=data.state_set[i],
                          allow_clip=True)
            _, var = gp._state_transit(action=data.action_set[i],
                                       state=data.state_set[i],
                                       required_var=True)
            print(res)
            print(data.new_state_set[i])
            print(np.sqrt(var))
            try:
                self.assertTrue(np.isclose(res,
                                           data.new_state_set[i], atol=1e-1).all())
                self.assertTrue(np.greater_equal(data.new_state_set[i], res - 10.0 * np.sqrt(var)).all())
                self.assertTrue(np.less_equal(data.new_state_set[i], res + 10.0 * np.sqrt(var)).all())
            except Exception as e:
                print(e)
                print(np.isclose(res, data.new_state_set[i], atol=1e-1).all())
                print(np.greater_equal(data.new_state_set[i], res - 10.0 * np.sqrt(var)).all())
                print(np.less_equal(data.new_state_set[i], res + 10.0 * np.sqrt(var)).all())
                raise e

        # do test
        print("gp test")
        data = get_some_samples(env=env, policy=policy, num=100, env_spec=env_spec)
        for i in range(len(data.state_set)):
            res = gp.step(action=data.action_set[i],
                          state=data.state_set[i],
                          allow_clip=True)
            _, var = gp._state_transit(action=data.action_set[i],
                                       state=data.state_set[i],
                                       required_var=True)
            print(res)
            print(data.new_state_set[i])
            print(np.sqrt(var))
            print('l1 loss {}'.format(np.linalg.norm(data.new_state_set[i] - res, 1)))
        return locals()



"""
A simple example to show how to build up an experiment with ddpg training and testing on Pendulum-v0
"""
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.algo.ddpg import DDPG
from baconian.algo.policy import DeterministicMLPPolicy
from baconian.core.agent import Agent
from baconian.algo.misc import EpsilonGreedy
from baconian.core.experiment import Experiment
from baconian.core.flow.train_test_flow import create_train_test_flow
from baconian.config.global_config import GlobalConfig
from baconian.core.status import get_global_status_collect
from baconian.common.schedules import PeriodicalEventSchedule
import baconian.common.log_data_loader as loader
from pathlib import Path


def task_fn():
    env = make('Pendulum-v0')
    name = 'demo_exp'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)

    mlp_q = MLPQValueFunction(env_spec=env_spec,
                              name_scope=name + '_mlp_q',
                              name=name + '_mlp_q',
                              mlp_config=[
                                  {
                                      "ACT": "RELU",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "1",
                                      "N_UNITS": 16,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "LINEAR",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "OUPTUT",
                                      "N_UNITS": 1,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  }
                              ])
    policy = DeterministicMLPPolicy(env_spec=env_spec,
                                    name_scope=name + '_mlp_policy',
                                    name=name + '_mlp_policy',
                                    mlp_config=[
                                        {
                                            "ACT": "RELU",
                                            "B_INIT_VALUE": 0.0,
                                            "NAME": "1",
                                            "N_UNITS": 16,
                                            "TYPE": "DENSE",
                                            "W_NORMAL_STDDEV": 0.03
                                        },
                                        {
                                            "ACT": "LINEAR",
                                            "B_INIT_VALUE": 0.0,
                                            "NAME": "OUPTUT",
                                            "N_UNITS": env_spec.flat_action_dim,
                                            "TYPE": "DENSE",
                                            "W_NORMAL_STDDEV": 0.03
                                        }
                                    ],
                                    reuse=False)

    ddpg = DDPG(
        env_spec=env_spec,
        config_or_config_dict={
            "REPLAY_BUFFER_SIZE": 10000,
            "GAMMA": 0.999,
            "CRITIC_LEARNING_RATE": 0.001,
            "ACTOR_LEARNING_RATE": 0.001,
            "DECAY": 0.5,
            "BATCH_SIZE": 50,
            "TRAIN_ITERATION": 1,
            "critic_clip_norm": 0.1,
            "actor_clip_norm": 0.1,
        },
        value_func=mlp_q,
        policy=policy,
        name=name + '_ddpg',
        replay_buffer=None
    )
    agent = Agent(env=env, env_spec=env_spec,
                  algo=ddpg,
                  algo_saving_scheduler=PeriodicalEventSchedule(
                      t_fn=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                      trigger_every_step=20,
                      after_t=10),
                  name=name + '_agent',
                  exploration_strategy=EpsilonGreedy(action_space=env_spec.action_space,
                                                     init_random_prob=0.5))

    flow = create_train_test_flow(
        test_every_sample_count=10,
        train_every_sample_count=10,
        start_test_after_sample_count=5,
        start_train_after_sample_count=5,
        train_func_and_args=(agent.train, (), dict()),
        test_func_and_args=(agent.test, (), dict(sample_count=1)),
        sample_func_and_args=(agent.sample, (), dict(sample_count=100,
                                                     env=agent.env,
                                                     store_flag=True))
    )

    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name
    )
    experiment.run()


from baconian.core.experiment_runner import *

GlobalConfig().set('DEFAULT_LOG_PATH', './log_path')
single_exp_runner(task_fn, del_if_log_path_existed=True)



import os
import sys

CURRENT_PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.append(CURRENT_PATH)
PAR_PATH = os.path.abspath(os.path.join(CURRENT_PATH, os.pardir))
sys.path.append(PAR_PATH)



import numpy as np
import tensorflow as tf
from baconian.config.global_config import GlobalConfig
from baconian.tf.util import create_new_tf_session
from baconian.test.tests.set_up.setup import TestWithAll


class TestTensorflowParameters(TestWithAll):
    def test_tf_param(self):
        param, _ = self.create_tf_parameters()
        param.init()
        param.save_snapshot()
        param.load_snapshot()

        para2, _ = self.create_tf_parameters(name='para2')
        para2.init()
        para2.copy_from(param)

        for key in param._source_config.required_key_dict.keys():
            if isinstance(param[key], tf.Tensor):
                continue
            if isinstance(param[key], np.ndarray):
                self.assertTrue(np.equal(param[key], para2[key]).all())
            else:
                self.assertEqual(param[key], para2[key])
                self.assertEqual(param(key), para2(key))
        for key in param._parameters.keys():
            if isinstance(param[key], tf.Tensor):
                continue

            if isinstance(param[key], np.ndarray):
                self.assertTrue(np.equal(param[key], para2[key]).all())
            else:
                self.assertEqual(param[key], para2[key])
                self.assertEqual(param(key), para2(key))

    def test_save_load(self):

        param, _ = self.create_tf_parameters('param')

        param.init()
        var_val = [self.sess.run(var) for var in param('tf_var_list')]

        param_other, _ = self.create_tf_parameters(name='other_param')
        param_other.init()

        for i in range(10):
            param.save(sess=self.sess,
                       save_path=GlobalConfig().DEFAULT_LOG_PATH + '/model',
                       global_step=i)

        if tf.get_default_session():
            sess = tf.get_default_session()
            sess.__exit__(None, None, None)
        tf.reset_default_graph()
        print('set tf device as {}'.format(self.default_id))
        import os
        os.environ["CUDA_VISIBLE_DEVICES"] = str(self.default_id)

        self.sess = create_new_tf_session()

        param2, _ = self.create_tf_parameters('param')
        param2.init()
        param2.load(path_to_model=GlobalConfig().DEFAULT_LOG_PATH + '/model', global_step=9)
        for var1, var2 in zip(var_val, param2('tf_var_list')):
            self.assertTrue(np.equal(var1, self.sess.run(var2)).all())



import unittest
from baconian.envs.gym_env import make
from baconian.core.core import EnvSpec
from baconian.algo.policy.normal_distribution_mlp import NormalDistributionMLPPolicy
from baconian.common.special import *
from baconian.test.tests.set_up.setup import TestTensorflowSetup


class TestNormalDistMLPPolicy(TestTensorflowSetup):
    def test_mlp_norm_dist_policy(self):
        env = make('Pendulum-v0')
        env.reset()
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)

        policy = NormalDistributionMLPPolicy(env_spec=env_spec,
                                             name='mlp_policy',
                                             name_scope='mlp_policy',
                                             mlp_config=[
                                                 {
                                                     "ACT": "RELU",
                                                     "B_INIT_VALUE": 0.0,
                                                     "NAME": "1",
                                                     "N_UNITS": 16,
                                                     "TYPE": "DENSE",
                                                     "W_NORMAL_STDDEV": 0.03
                                                 },
                                                 {
                                                     "ACT": "LINEAR",
                                                     "B_INIT_VALUE": 0.0,
                                                     "NAME": "OUPTUT",
                                                     "N_UNITS": env_spec.flat_action_dim,
                                                     "TYPE": "DENSE",
                                                     "W_NORMAL_STDDEV": 0.03
                                                 }
                                             ],
                                             output_high=None,
                                             output_low=None,
                                             output_norm=None,
                                             input_norm=None,
                                             reuse=False)
        policy.init()
        dist_info = policy.get_dist_info()
        self.assertTrue(np.equal(dist_info[0]['shape'], policy.mean_output.shape.as_list()).all())
        self.assertTrue(np.equal(dist_info[1]['shape'], policy.logvar_output.shape.as_list()).all())
        for _ in range(10):
            ac = policy.forward(obs=env.observation_space.sample())
            self.assertTrue(env.action_space.contains(ac[0]))
        p2 = policy.make_copy(name='test',
                              name_scope='mlp_policy_2',
                              reuse=False)
        p2.init()
        self.assertGreater(len(policy.parameters('tf_var_list')), 0)
        self.assertGreater(len(p2.parameters('tf_var_list')), 0)
        for var1, var2 in zip(policy.parameters('tf_var_list'), p2.parameters('tf_var_list')):
            self.assertEqual(var1.shape, var2.shape)
            self.assertNotEqual(id(var1), id(var2))

        p3 = policy.make_copy(name='mlp_policy_ttt',
                              name_scope='mlp_policy',
                              reuse=True)
        p3.init()
        self.assertGreater(len(p3.parameters('tf_var_list')), 0)
        for var1, var2 in zip(policy.parameters('tf_var_list'), p3.parameters('tf_var_list')):
            self.assertEqual(var1.shape, var2.shape)
            self.assertEqual(id(var1), id(var2))

        # policy.copy_from(p2)]
        res_not_true = []
        for var1, var2, var3 in zip(policy.parameters('tf_var_list'), p2.parameters('tf_var_list'),
                                    p3.parameters('tf_var_list')):
            re1, re2, re3 = self.sess.run([var1, var2, var3])
            res_not_true.append(np.isclose(re1, re2).all())
            res_not_true.append(np.isclose(re3, re2).all())
            self.assertTrue(np.isclose(re1, re3).all())
        self.assertFalse(np.array(res_not_true).all())

        policy.copy_from(p2)

        for var1, var2, var3 in zip(policy.parameters('tf_var_list'), p2.parameters('tf_var_list'),
                                    p3.parameters('tf_var_list')):
            re1, re2, re3 = self.sess.run([var1, var2, var3])
            self.assertTrue(np.isclose(re1, re3).all())
            self.assertTrue(np.isclose(re2, re3).all())
            self.assertTrue(np.isclose(re1, re2).all())

    def test_func(self):
        env = make('Pendulum-v0')
        env.reset()
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)

        policy = NormalDistributionMLPPolicy(env_spec=env_spec,
                                             name='mlp_policy',
                                             name_scope='mlp_policy',
                                             mlp_config=[
                                                 {
                                                     "ACT": "RELU",
                                                     "B_INIT_VALUE": 0.0,
                                                     "NAME": "1",
                                                     "N_UNITS": 16,
                                                     "TYPE": "DENSE",
                                                     "W_NORMAL_STDDEV": 0.03
                                                 },
                                                 {
                                                     "ACT": "LINEAR",
                                                     "B_INIT_VALUE": 0.0,
                                                     "NAME": "OUPTUT",
                                                     "N_UNITS": env_spec.flat_action_dim,
                                                     "TYPE": "DENSE",
                                                     "W_NORMAL_STDDEV": 0.03
                                                 }
                                             ],
                                             output_high=None,
                                             output_low=None,
                                             output_norm=None,
                                             input_norm=None,
                                             reuse=False)
        policy.init()
        print(
            policy.compute_dist_info(name='entropy',
                                     feed_dict={
                                         policy.state_input: make_batch(env_spec.obs_space.sample(),
                                                                        original_shape=env_spec.obs_shape)}))
        print(
            policy.compute_dist_info(name='prob',
                                     value=env_spec.action_space.sample(),
                                     feed_dict={
                                         policy.state_input: make_batch(env_spec.obs_space.sample(),
                                                                        original_shape=env_spec.obs_shape),
                                         policy.action_input: make_batch(env_spec.action_space.sample(),
                                                                         original_shape=env_spec.action_shape)}))
        new_policy = policy.make_copy(
            reuse=False,
            name='new_p',
            name_scope='mlp_policy_2',

        )
        new_policy.init()
        for var1, var2 in zip(policy.parameters('tf_var_list'), new_policy.parameters('tf_var_list')):
            print(var1.name)
            print(var2.name)
            self.assertNotEqual(var1.name, var2.name)
            self.assertNotEqual(id(var1), id(var2))
        obs1 = make_batch(env_spec.obs_space.sample(),
                          original_shape=env_spec.obs_shape,
                          )
        obs2 = make_batch(env_spec.obs_space.sample(),
                          original_shape=env_spec.obs_shape)
        kl1 = policy.compute_dist_info(name='kl', other=new_policy, feed_dict={
            policy.state_input: obs1,
            new_policy.state_input: obs2
        })
        kl2 = self.sess.run(policy.kl(other=new_policy), feed_dict={
            policy.state_input: obs1,
            new_policy.state_input: obs2
        })
        self.assertTrue(np.isclose(kl1, kl2).all())


if __name__ == '__main__':
    unittest.main()



from baconian.test.tests.set_up.setup import TestWithAll
from baconian.common.logging import Logger, ConsoleLogger, Recorder, record_return_decorator
import numpy as np
from baconian.core.core import Basic, EnvSpec
from baconian.algo.dqn import DQN
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.core.agent import Agent


class Foo(Basic):
    def __init__(self, name='foo'):
        super().__init__(name=name)
        self.loss = 1.0
        self.recorder = Recorder(flush_by_split_status=False, default_obj=self)

    def get_status(self):
        return dict(x=1)

    def get_val(self):
        return np.random.random()

    @record_return_decorator(which_recorder='self')
    def get_by_return(self, res, num=2, *args, **kwargs):
        return dict(val=res * num, val2=res)

    @property
    def name(self):
        return 'foo'


class TestLogger(TestWithAll):
    def test_register(self):
        obj = Foo()
        a = Recorder(flush_by_split_status=False, default_obj=obj)
        a.register_logging_attribute_by_record(obj=obj, attr_name='val', get_method=lambda x: x['obj'].get_val(),
                                               static_flag=False)
        a.register_logging_attribute_by_record(obj=obj, attr_name='loss', static_flag=True)
        a.record()
        print(a._obj_log)
        self.assertTrue('val' in a._obj_log[obj])
        self.assertTrue('loss' in a._obj_log[obj])
        obj.loss = 10.0
        a.record()

        b = Recorder(flush_by_split_status=False, default_obj=obj)
        b.register_logging_attribute_by_record(obj=obj, attr_name='val', get_method=lambda x: x['obj'].get_val(),
                                               static_flag=False)
        b.register_logging_attribute_by_record(obj=obj, attr_name='loss', static_flag=True)

        b.record()
        self.assertTrue('val' in b._obj_log[obj])
        self.assertTrue('loss' in b._obj_log[obj])
        obj.loss = 10.0
        b.record()
        self.assertTrue(b._obj_log is not a._obj_log)
        self.assertTrue(b._registered_log_attr_by_get_dict is not a._registered_log_attr_by_get_dict)

    def test_return_record(self):
        obj = Foo(name='foo')
        obj.get_by_return(res=10, num=2)
        obj.get_by_return(res=1, num=2)
        obj.get_by_return(res=2, num=4)
        print(obj.recorder._obj_log)
        self.assertEqual(len(obj.recorder._obj_log), 1)
        self.assertTrue(obj in obj.recorder._obj_log)
        self.assertTrue('val' in obj.recorder._obj_log[obj])
        self.assertTrue(len(obj.recorder._obj_log[obj]['val']) == 3)
        self.assertTrue(obj.recorder._obj_log[obj]['val'][0]['log_val'] == 20)
        self.assertTrue(obj.recorder._obj_log[obj]['val'][1]['log_val'] == 2)
        self.assertTrue(obj.recorder._obj_log[obj]['val'][2]['log_val'] == 8)

        self.assertTrue('val2' in obj.recorder._obj_log[obj])
        self.assertTrue(len(obj.recorder._obj_log[obj]['val2']) == 3)
        self.assertTrue(obj.recorder._obj_log[obj]['val2'][0]['log_val'] == 10)
        self.assertTrue(obj.recorder._obj_log[obj]['val2'][1]['log_val'] == 1)
        self.assertTrue(obj.recorder._obj_log[obj]['val2'][2]['log_val'] == 2)

        obj = Foo(name='foo2')
        obj.get_by_return(res=10, num=2)
        obj.get_by_return(res=1, num=2)
        obj.get_by_return(res=2, num=4)
        print(obj.recorder._obj_log)
        self.assertTrue(obj in obj.recorder._obj_log)
        self.assertTrue('val' in obj.recorder._obj_log[obj])
        self.assertTrue(len(obj.recorder._obj_log[obj]['val']) == 3)
        self.assertTrue(obj.recorder._obj_log[obj]['val'][0]['log_val'] == 20)
        self.assertTrue(obj.recorder._obj_log[obj]['val'][1]['log_val'] == 2)
        self.assertTrue(obj.recorder._obj_log[obj]['val'][2]['log_val'] == 8)

        self.assertTrue('val2' in obj.recorder._obj_log[obj])
        self.assertTrue(len(obj.recorder._obj_log[obj]['val2']) == 3)
        self.assertTrue(obj.recorder._obj_log[obj]['val2'][0]['log_val'] == 10)
        self.assertTrue(obj.recorder._obj_log[obj]['val2'][1]['log_val'] == 1)
        self.assertTrue(obj.recorder._obj_log[obj]['val2'][2]['log_val'] == 2)


class TesTLoggerWithDQN(TestWithAll):

    def test_integration_with_dqn(self):
        env = make('Acrobot-v1')
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)

        mlp_q = MLPQValueFunction(env_spec=env_spec,
                                  name='mlp_q',
                                  name_scope='mlp_q',
                                  mlp_config=[
                                      {
                                          "ACT": "RELU",
                                          "B_INIT_VALUE": 0.0,
                                          "NAME": "1",
                                          "N_UNITS": 16,
                                          "TYPE": "DENSE",
                                          "W_NORMAL_STDDEV": 0.03
                                      },
                                      {
                                          "ACT": "LINEAR",
                                          "B_INIT_VALUE": 0.0,
                                          "NAME": "OUPTUT",
                                          "N_UNITS": 1,
                                          "TYPE": "DENSE",
                                          "W_NORMAL_STDDEV": 0.03
                                      }
                                  ])
        dqn = DQN(env_spec=env_spec,
                  name='dqn_test',
                  config_or_config_dict=dict(REPLAY_BUFFER_SIZE=1000,
                                             GAMMA=0.99,
                                             BATCH_SIZE=10,
                                             LEARNING_RATE=0.001,
                                             TRAIN_ITERATION=1,
                                             DECAY=0.5),
                  value_func=mlp_q)
        agent = Agent(env=env, env_spec=env_spec,
                      algo=dqn,
                      name='agent')
        agent.init()
        # dqn.init()
        st = env.reset()
        from baconian.common.sampler.sample_data import TransitionData
        a = TransitionData(env_spec)
        res = []
        agent.sample(env=env,
                     sample_count=100,
                     in_which_status='TRAIN',
                     store_flag=True,
                     sample_type='transition')
        agent.sample(env=env,
                     sample_count=100,
                     in_which_status='TRAIN',
                     store_flag=True,
                     sample_type='transition')
        res.append(dqn.train(batch_data=a, train_iter=10, sess=None, update_target=True)['average_loss'])
        res.append(dqn.train(batch_data=None, train_iter=10, sess=None, update_target=True)['average_loss'])
        self.assertTrue(dqn in dqn.recorder._obj_log)
        self.assertTrue('average_loss' in dqn.recorder._obj_log[dqn])
        self.assertTrue(len(dqn.recorder._obj_log[dqn]['average_loss']) == 2)
        self.assertTrue(
            np.equal(np.array(res), [x['log_val'] for x in dqn.recorder._obj_log[dqn]['average_loss']]).all())

        self.assertTrue(len(Logger()._registered_recorders) > 0)
        self.assertTrue(dqn.recorder in Logger()._registered_recorders)
        res = dqn.recorder.get_log(attr_name='average_loss', filter_by_status=dict())
        self.assertEqual(len(res), 2)
        res = agent.recorder.get_log(attr_name='sum_reward', filter_by_status={'status': 'TRAIN'})
        self.assertEqual(len(res), 2)
        res = agent.recorder.get_log(attr_name='sum_reward', filter_by_status={'status': 'TEST'})
        self.assertEqual(len(res), 0)
        Logger().flush_recorder()

    def test_console_logger(self):
        self.assertTrue(ConsoleLogger().inited_flag)
        logger = ConsoleLogger()
        self.assertTrue(logger.inited_flag)
        logger.print('info', 'this is for test %s', 'args')
        logger.print('info', 'this is for test {}'.format('args'))

        logger2 = ConsoleLogger()
        self.assertEqual(id(logger), id(logger2))
        logger.flush()


if __name__ == '__main__':
    import unittest

    unittest.main()



"""
Dyna benchmark on Pendulum
"""

from baconian.benchmark.dyna_benchmark.pendulum_conf import PENDULUM_BENCHMARK_CONFIG_DICT as exp_config
from baconian.common.noise import *
from baconian.common.schedules import *
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.algo.ddpg import DDPG
from baconian.algo.policy import DeterministicMLPPolicy
from baconian.core.agent import Agent
from baconian.core.experiment import Experiment
from baconian.config.global_config import GlobalConfig
from baconian.core.status import get_global_status_collect
from baconian.algo.dynamics.mlp_dynamics_model import ContinuousMLPGlobalDynamicsModel
from baconian.algo.dyna import Dyna
from baconian.algo.dynamics.terminal_func.terminal_func import FixedEpisodeLengthTerminalFunc
from baconian.core.flow.dyna_flow import DynaFlow
from baconian.envs.envs_reward_func import REWARD_FUNC_DICT


def pendulum_task_fn():
    GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT',
                       exp_config['DEFAULT_EXPERIMENT_END_POINT'])

    env = make('Pendulum-v0')
    name = 'benchmark'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)

    mlp_q = MLPQValueFunction(env_spec=env_spec,
                              name_scope=name + '_mlp_q',
                              name=name + '_mlp_q',
                              **exp_config['MLPQValueFunction'])
    policy = DeterministicMLPPolicy(env_spec=env_spec,
                                    name_scope=name + '_mlp_policy',
                                    name=name + '_mlp_policy',
                                    output_low=env_spec.action_space.low,
                                    output_high=env_spec.action_space.high,
                                    **exp_config['DeterministicMLPPolicy'],
                                    reuse=False)

    ddpg = DDPG(
        env_spec=env_spec,
        policy=policy,
        value_func=mlp_q,
        name=name + '_ddpg',
        **exp_config['DDPG']
    )

    mlp_dyna = ContinuousMLPGlobalDynamicsModel(
        env_spec=env_spec,
        name_scope=name + '_mlp_dyna',
        name=name + '_mlp_dyna',
        **exp_config['DynamicsModel']
    )
    algo = Dyna(env_spec=env_spec,
                name=name + '_dyna_algo',
                model_free_algo=ddpg,
                dynamics_model=mlp_dyna,
                config_or_config_dict=dict(
                    dynamics_model_train_iter=10,
                    model_free_algo_train_iter=10
                ))
    algo.set_terminal_reward_function_for_dynamics_env(
        terminal_func=FixedEpisodeLengthTerminalFunc(max_step_length=env.unwrapped._max_episode_steps,
                                                     step_count_fn=algo.dynamics_env.total_step_count_fn),
        reward_func=REWARD_FUNC_DICT['Pendulum-v0']())
    agent = Agent(env=env, env_spec=env_spec,
                  algo=algo,
                  exploration_strategy=None,
                  noise_adder=AgentActionNoiseWrapper(noise=OrnsteinUhlenbeckActionNoise(np.zeros(1, ), 0.15),
                                                      noise_weight_scheduler=ConstantScheduler(value=0.3),
                                                      action_weight_scheduler=ConstantScheduler(value=1.0)),
                  name=name + '_agent')

    flow = DynaFlow(
        train_sample_count_func=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
        config_or_config_dict=exp_config['DynaFlow'],
        func_dict={
            'train_algo': {'func': agent.train,
                           'args': list(),
                           'kwargs': dict(state='state_agent_training')},
            'train_algo_from_synthesized_data': {'func': agent.train,
                                                 'args': list(),
                                                 # TODO use a decomposed way to represetn the state
                                                 #  e.g., TRAIN:AGENT:CYBER
                                                 'kwargs': dict(state='state_agent_training', train_iter=1)},
            'train_dynamics': {'func': agent.train,
                               'args': list(),
                               'kwargs': dict(state='state_dynamics_training')},
            'test_algo': {'func': agent.test,
                          'args': list(),
                          'kwargs': dict(sample_count=1)},
            'test_dynamics': {'func': agent.algo.test_dynamics,
                              'args': list(),
                              'kwargs': dict(sample_count=10, env=env)},
            'sample_from_real_env': {'func': agent.sample,
                                     'args': list(),
                                     'kwargs': dict(sample_count=20,
                                                    env=agent.env,
                                                    sample_type='transition',
                                                    in_which_status='TRAIN',
                                                    store_flag=True)},
            'sample_from_dynamics_env': {'func': agent.sample,
                                         'args': list(),
                                         'kwargs': dict(sample_count=20,
                                                        sample_type='transition',
                                                        env=agent.algo.dynamics_env,
                                                        in_which_status='TRAIN',
                                                        store_flag=False)}
        }
    )

    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name
    )
    experiment.run()



"""
PPO benchmark on Swimmer
"""
from baconian.benchmark.ppo_benchmark.swimmer_conf import *
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func import MLPVValueFunc
from baconian.algo.ppo import PPO
from baconian.algo.policy.normal_distribution_mlp import NormalDistributionMLPPolicy
from baconian.core.agent import Agent
from baconian.core.experiment import Experiment
from baconian.core.flow.train_test_flow import TrainTestFlow
from baconian.config.global_config import GlobalConfig
from baconian.core.status import get_global_status_collect


def swimmer_task_fn():
    exp_config = SWIMMER_BENCHMARK_CONFIG_DICT
    GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT',
                       exp_config['DEFAULT_EXPERIMENT_END_POINT'])

    env = make('Swimmer-v2')
    name = 'benchmark'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)

    mlp_v = MLPVValueFunc(env_spec=env_spec,
                          name_scope=name + 'mlp_v',
                          name=name + 'mlp_v',
                          **exp_config['MLP_V'])
    policy = NormalDistributionMLPPolicy(env_spec=env_spec,
                                         name_scope=name + 'mlp_policy',
                                         name=name + 'mlp_policy',
                                         **exp_config['POLICY'],
                                         output_low=env_spec.action_space.low,
                                         output_high=env_spec.action_space.high,
                                         reuse=False)

    ppo = PPO(
        env_spec=env_spec,
        **exp_config['PPO'],
        value_func=mlp_v,
        stochastic_policy=policy,
        name=name + 'ppo'
    )
    agent = Agent(env=env, env_spec=env_spec,
                  algo=ppo,
                  exploration_strategy=None,
                  noise_adder=None,
                  name=name + '_agent')

    flow = TrainTestFlow(train_sample_count_func=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                         config_or_config_dict=exp_config['TrainTestFlow']['config_or_config_dict'],
                         func_dict={
                             'test': {'func': agent.test,
                                      'args': list(),
                                      'kwargs': dict(sample_count=exp_config['TrainTestFlow']['TEST_SAMPLES_COUNT']),
                                      },
                             'train': {'func': agent.train,
                                       'args': list(),
                                       'kwargs': dict(),
                                       },
                             'sample': {'func': agent.sample,
                                        'args': list(),
                                        'kwargs': dict(sample_count=exp_config['TrainTestFlow']['TRAIN_SAMPLES_COUNT'],
                                                       env=agent.env,
                                                       sample_type='trajectory',
                                                       in_which_status='TRAIN',
                                                       store_flag=True),
                                        },
                         })

    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name
    )
    experiment.run()

